import json
import os
import re
import shutil
from pathlib import Path

import pytest
from eth_utils import to_hex
from ethpm_types import Compiler, ContractType, PackageManifest, Source
from ethpm_types.manifest import PackageName

from ape import Project
from ape.contracts import ContractContainer
from ape.exceptions import ConfigError, ProjectError
from ape.logging import LogLevel
from ape.managers.project import MultiProject
from ape.utils import create_tempdir
from ape_pm.project import BrownieProject, FoundryProject
from tests.conftest import skip_if_plugin_installed


@pytest.fixture
def tmp_project(with_dependencies_project_path):
    real_project = Project(with_dependencies_project_path)
    # Copies contracts and stuff into a temp folder
    # and returns a project around the temp folder.
    with real_project.isolate_in_tempdir() as tmp_project:
        yield tmp_project


@pytest.fixture
def contract_type():
    return make_contract("FooContractFromManifest")


@pytest.fixture
def manifest(contract_type):
    return make_manifest(contract_type)


@pytest.fixture
def contract_block_hash(eth_tester_provider, vyper_contract_instance):
    block_number = vyper_contract_instance.creation_metadata.block
    return to_hex(eth_tester_provider.get_block(block_number).hash)


@pytest.fixture
def project_from_manifest(manifest):
    return Project.from_manifest(manifest)


@pytest.fixture(scope="module")
def foundry_toml():
    return """
[profile.default]
src = 'src'
out = 'out'
libs = ['lib']
solc = "0.8.18"
evm_version = 'cancun'
via_ir = true

remappings = [
    'forge-std/=lib/forge-std/src/',
    '@openzeppelin/=lib/openzeppelin-contracts/',
]
""".lstrip()


def make_contract(name: str = "test") -> ContractType:
    return ContractType.model_validate(
        {
            "contractName": name,
            "sourceId": f"contracts/{name}.json",
            "abi": [],
        }
    )


def make_manifest(*contracts: ContractType, include_contract_type: bool = True) -> PackageManifest:
    sources = {
        ct.source_id: Source(content=ct.model_dump_json(by_alias=True, mode="json"))
        for ct in contracts
    }
    model: dict = {"sources": sources}
    if include_contract_type:
        contract_types = {c.name: c for c in contracts}
        model["contractTypes"] = contract_types

    return PackageManifest.model_validate(model)


def test_path(project):
    assert project.path is not None


def test_path_configured(project):
    """
    Simulating package structures like snekmate.
    """
    madeup_name = "snakemate"
    with create_tempdir(name=madeup_name) as temp_dir:
        subdir = temp_dir / "src"
        contracts_folder = subdir / madeup_name
        contracts_folder.mkdir(parents=True)
        contract = contracts_folder / "snake.json"
        abi = [{"name": "foo", "type": "fallback", "stateMutability": "nonpayable"}]
        contract.write_text(json.dumps(abi), encoding="utf8")

        snakemate = Project(
            temp_dir, config_override={"base_path": "src", "contracts_folder": madeup_name}
        )
        assert snakemate.name == madeup_name
        assert snakemate.path == subdir
        assert snakemate.contracts_folder == contracts_folder

        # The repr should show `/snakemate` and not `/snakemate/src/`.
        assert re.match(r"<ProjectManager [\w|/]*/snakemate>", repr(snakemate))

        actual = snakemate.load_contracts()
        assert "snake" in actual
        assert actual["snake"].source_id == f"{madeup_name}/snake.json"


def test_name(project):
    assert project.name == project.path.name


def test_name_from_config(project):
    with project.temp_config(name="foo-bar"):
        assert project.name == "foo-bar"


def test_repr():
    with create_tempdir() as tmpdir:
        project = Project(tmpdir)
        actual = repr(project)

    expected = f"<ProjectManager {tmpdir}>"
    assert actual == expected


@pytest.mark.parametrize("name", ("contracts", "sources"))
def test_contracts_folder_from_config(project, name):
    with project.temp_config(contracts_folder=name):
        assert project.contracts_folder == project.path / name


def test_contracts_folder_same_as_root_path(project):
    with project.temp_config(contracts_folder="."):
        assert project.contracts_folder == project.path


def test_contracts_folder_deduced(tmp_project):
    new_project_path = tmp_project.path / "new"
    new_project_path.mkdir()
    contracts_folder = new_project_path / "sources"
    contracts_folder.mkdir()
    contract = contracts_folder / "tryme.json"
    abi = [{"name": "foo", "type": "fallback", "stateMutability": "nonpayable"}]
    contract.write_text(json.dumps(abi), encoding="utf8")
    new_project = Project(new_project_path)
    actual = new_project.contracts_folder
    assert actual == contracts_folder


def test_reconfigure(project):
    project.reconfigure(compile={"exclude": ["first", "second"]})
    assert {"first", "second"}.issubset(set(project.config.compile.exclude))


def test_isolate_in_tempdir(project):
    # Purposely not using `tmp_project` fixture.
    with project.isolate_in_tempdir() as tmp_project:
        assert tmp_project.path != project.path
        assert tmp_project.in_tempdir


def test_isolate_in_tempdir_does_not_alter_sources(project):
    # First, create a bad source.
    with project.temp_config(contracts_folder="build"):
        new_src = project.contracts_folder / "newsource.json"
        new_src.parent.mkdir(exist_ok=True, parents=True)
        new_src.write_text("this is not json, oops")
        project.sources.refresh()  # Only need to be called when run with other tests.

        try:
            with project.isolate_in_tempdir() as tmp_project:
                # The new (bad) source should be in the temp project.
                actual = {**(tmp_project.manifest.sources or {})}
        finally:
            new_src.unlink()
            project.sources.refresh()

        # Ensure "newsource" did not persist in the in-memory manifest.
        assert "build/newsource.json" in actual
        assert "build/newsource.json" not in (project.manifest.sources or {})


def test_in_tempdir(project, tmp_project):
    assert not project.in_tempdir
    assert tmp_project.in_tempdir


def test_getattr(tmp_project):
    actual = tmp_project.Other
    assert type(actual) is ContractContainer


def test_getattr_not_exists(tmp_project):
    expected = (
        r"'LocalProject' object has no attribute 'nope'\. Also checked extra\(s\) 'contracts'\."
    )
    with pytest.raises(AttributeError, match=expected) as err:
        _ = tmp_project.nope

    # Was the case where the last entry was from Ape's basemodel stuff.
    # Now, it points at the project manager last.
    assert "ape/managers/project.py:" in repr(err.traceback[-1])


def test_getattr_detects_changes(tmp_project):
    source_id = tmp_project.Other.contract_type.source_id
    new_abi = {
        "inputs": [],
        "name": "retrieve",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    }
    content = json.dumps([new_abi])
    path = tmp_project.sources.lookup(source_id)
    assert path
    path.unlink(missing_ok=True)
    path.write_text(content, encoding="utf8")
    # Should have re-compiled.
    contract = tmp_project.Other
    assert "retrieve" in contract.contract_type.methods


def test_getattr_empty_contract(tmp_project):
    """
    Tests against a condition where would infinitely compile.
    """
    source_id = tmp_project.Other.contract_type.source_id
    path = tmp_project.sources.lookup(source_id)
    path.unlink(missing_ok=True)
    path.write_text("", encoding="utf8")
    # Should have re-compiled.
    contract = tmp_project.Other
    assert not contract.contract_type.methods


@skip_if_plugin_installed("vyper", "solidity")
def test_getattr_same_name_as_source_file(project_with_source_files_contract):
    expected = (
        r"'LocalProject' object has no attribute 'ContractA'\. "
        r"Also checked extra\(s\) 'contracts'\. "
        r"However, there is a source file named 'ContractA\.sol'\. "
        r"This file may not be compiling \(see error above\), "
        r"or maybe you meant to reference a contract name from this source file\? "
        r"Else, could it be from one of the missing compilers for extensions: .*"
    )
    with pytest.raises(AttributeError, match=expected):
        _ = project_with_source_files_contract.ContractA


@pytest.mark.parametrize("iypthon_attr_name", ("_repr_mimebundle_", "_ipython_display_"))
def test_getattr_ipython(tmp_project, iypthon_attr_name):
    # Remove contract types, if there for some reason is any.
    tmp_project.manifest.contract_types = {}
    getattr(tmp_project, iypthon_attr_name)
    # Prove it did not compile looking for these names.
    assert not tmp_project.manifest.contract_types


def test_getattr_ipython_canary_check(tmp_project):
    tmp_project.manifest.contract_types = {}

    with pytest.raises(AttributeError):
        _ = tmp_project._ipython_canary_method_should_not_exist_

    # Prove it did not compile looking for this.
    assert not tmp_project.manifest.contract_types


def test_getitem(tmp_project):
    actual = tmp_project["Project"]
    assert type(actual) is ContractContainer


def test_meta(project):
    meta_config = {
        "meta": {
            "authors": ["Apealicious Jones"],
            "license": "MIT",
            "description": "Zoologist meme protocol",
            "keywords": ["Indiana", "Knight's Templar"],
            "links": {"apeworx.io": "https://apeworx.io"},
        }
    }
    with project.temp_config(**meta_config):
        assert project.meta.authors == ["Apealicious Jones"]
        assert project.meta.license == "MIT"
        assert project.meta.description == "Zoologist meme protocol"
        assert project.meta.keywords == ["Indiana", "Knight's Templar"]
        assert len(project.meta.links) == 1
        assert f"{project.meta.links['apeworx.io']}" == "https://apeworx.io/"


def test_extract_manifest(tmp_project, vyper_contract_instance):
    contract_type = vyper_contract_instance.contract_type
    tmp_project.manifest.contract_types = {contract_type.name: contract_type}
    tmp_project.deployments.track(vyper_contract_instance, allow_dev=True)

    manifest = tmp_project.extract_manifest()
    assert type(manifest) is PackageManifest
    assert manifest.meta == tmp_project.meta
    assert PackageName("manifest-dependency") in (manifest.dependencies or {})
    bip122_chain_id = to_hex(tmp_project.provider.get_block(0).hash)
    expected_uri = f"blockchain://{bip122_chain_id[2:]}"
    for key in manifest.deployments or {}:
        if key.startswith(expected_uri):
            return

    assert False, "Failed to find expected deployment URI"


def test_extract_manifest_when_sources_missing(tmp_project):
    """
    Show that if a source is missing, it is OK. This happens when changing branches
    after compiling and sources are only present on one of the branches.
    """
    contract = make_contract("notreallyhere")
    tmp_project.manifest.contract_types = {contract.name: contract}
    manifest = tmp_project.extract_manifest()

    # Source is skipped because missing.
    assert "notreallyhere" not in manifest.contract_types


def test_extract_manifest_excludes_cache(tmp_project):
    cachefile = tmp_project.contracts_folder / ".cache" / "CacheFile.json"
    cachefile2 = tmp_project.contracts_folder / ".cache" / "subdir" / "Cache2.json"
    cachefile2.parent.mkdir(parents=True)
    cachefile.write_text("Doesn't matter", encoding="utf8")
    cachefile2.write_text("Doesn't matter", encoding="utf8")
    manifest = tmp_project.extract_manifest()
    assert isinstance(manifest, PackageManifest)
    assert ".cache/CacheFile.json" not in (manifest.sources or {})
    assert ".cache/subdir/CacheFile.json" not in (manifest.sources or {})


def test_extract_manifest_compiles(tmp_project):
    tmp_project.manifest.contract_types = {}  # Not compiled.
    actual = tmp_project.extract_manifest()
    assert actual.contract_types  # Fails if empty


def test_extract_manifest_from_manifest_project(project_from_manifest):
    project_from_manifest.manifest.contract_types = {}  # Not compiled.
    manifest = project_from_manifest.extract_manifest()
    assert "FooContractFromManifest" in manifest.contract_types


def test_exclusions(tmp_project):
    exclusions = ["Other.json", "*Excl*"]
    exclude_config = {"compile": {"exclude": exclusions}}
    with tmp_project.temp_config(**exclude_config):
        for exclusion in exclusions:
            assert exclusion in tmp_project.exclusions


def test_update_manifest(tmp_project):
    compiler = Compiler(name="comp", version="1.0.0", contractTypes=["foo.txt"])
    tmp_project.update_manifest(compilers=[compiler])
    actual = tmp_project.manifest.compilers
    assert actual == [compiler]

    tmp_project.update_manifest(name="test", version="1.0.0")
    assert tmp_project.manifest.name == "test"
    assert tmp_project.manifest.version == "1.0.0"

    # The compilers should not have changed.
    actual = tmp_project.manifest.compilers
    assert actual == [compiler]


def test_load_contracts(tmp_project):
    contracts = tmp_project.load_contracts()
    assert tmp_project.manifest_path.is_file()
    assert len(contracts) > 0
    contracts_forced = tmp_project.load_contracts(use_cache=False)
    assert contracts_forced == contracts

    # Delete a file and ensure it is does not show up in dict.
    contract_to_rm = next(iter(contracts.values()))
    (tmp_project.path / contract_to_rm.source_id).unlink()
    contracts = tmp_project.load_contracts()
    assert contract_to_rm.name not in contracts


def test_load_contracts_detect_change(tmp_project, ape_caplog):
    path = tmp_project.contracts_folder / "Other.json"
    content = path.read_text()
    assert "foo" in content, "Test setup failed. Unexpected file content."

    # Must be compiled first.
    with ape_caplog.at_level(LogLevel.INFO):
        contracts = tmp_project.load_contracts(use_cache=False)
        assert "Other" in contracts
        ape_caplog.assert_last_log("Compiling")

        ape_caplog.clear()

        # No logs as it doesn't need to re-compile.
        tmp_project.load_contracts()
        assert not ape_caplog.head

        # Make a change to the file.
        new_content = content.replace("foo", "bar")
        assert "bar" in new_content, "Test setup failed. Unexpected file content."
        path.unlink()
        path.write_text(new_content, encoding="utf8")

        # Prove re-compiles.
        contracts = tmp_project.load_contracts()
        assert "Other" in contracts
        ape_caplog.assert_last_log("Compiling")


def test_load_contracts_after_deleting_same_named_contract(tmp_project, compilers, mock_compiler):
    """
    Tests against a scenario where you:

    1. Add and compile a contract
    2. Delete that contract
    3. Add a new contract with same name somewhere else

    Test such that we are able to compile successfully and not get a misleading
    collision error from deleted files.
    """
    init_contract = tmp_project.contracts_folder / "foo.__mock__"
    init_contract.write_text("LALA", encoding="utf8")
    compilers.registered_compilers[".__mock__"] = mock_compiler
    result = tmp_project.load_contracts()
    assert "foo" in result

    # Goodbye.
    init_contract.unlink()

    # Since we are changing files mid-session, we need to refresh the project.
    # Typically, users don't have to do this.
    tmp_project.refresh_sources()

    result = tmp_project.load_contracts()
    assert "foo" not in result  # Was deleted.
    # Also ensure it is gone from paths.
    assert "foo.__mock__" not in [x.name for x in tmp_project.sources.paths]

    # Create a new contract with the same name.
    new_contract = tmp_project.contracts_folder / "bar.__mock__"
    new_contract.write_text("BAZ", encoding="utf8")
    tmp_project.refresh_sources()

    mock_compiler.overrides = {"contractName": "foo"}
    result = tmp_project.load_contracts()
    assert "foo" in result


def test_load_contracts_output_abi(tmp_project):
    cfg = {"output_extra": ["ABI"]}
    with tmp_project.temp_config(compile=cfg):
        _ = tmp_project.load_contracts(use_cache=False)
        abi_folder = tmp_project.manifest_path.parent / "abi"
        assert abi_folder.is_dir()
        files = [x for x in abi_folder.iterdir()]
        assert len(files) > 0
        for file in files:
            assert file.suffix == ".json"

        # Ensure is usable.
        data = json.loads(file.read_text(encoding="utf8"))
        assert isinstance(data, list)
        assert len(data) >= 1
        # There was bug where this was a str.
        assert isinstance(data[0], dict)


def test_load_contracts_use_cache(mocker, tmp_project):
    """
    Showing the 'use_cache=bool' kwarg works.
    """
    compile_spy = mocker.spy(tmp_project.contracts, "_compile")

    tmp_project.manifest.contract_types = {}  # Force initial compile.
    contracts = tmp_project.load_contracts(use_cache=True)
    assert "Other" in contracts  # Other.json contract.
    assert "Project" in contracts  # Project.json contract.
    assert compile_spy.call_args_list[-1][-1]["use_cache"] is True

    # Show they get added to the manifest.
    assert "Other" in tmp_project.manifest.contract_types
    assert "Project" in tmp_project.manifest.contract_types

    # Showe we can use the cache again (no compiling!)
    contracts = tmp_project.load_contracts(use_cache=True)
    assert "Other" in contracts  # Other.json contract.
    assert "Project" in contracts  # Project.json contract.
    assert compile_spy.call_args_list[-1][-1]["use_cache"] is True

    # Show force-recompiles.
    contracts = tmp_project.load_contracts(use_cache=False)
    assert "Other" in contracts  # Other.json contract.
    assert "Project" in contracts  # Project.json contract.
    assert compile_spy.call_args_list[-1][-1]["use_cache"] is False


def test_manifest_path(tmp_project):
    assert tmp_project.manifest_path == tmp_project.path / ".build" / "__local__.json"


def test_clean(tmp_project):
    tmp_project.load_contracts()
    assert tmp_project.manifest_path.is_file()

    tmp_project.clean()
    assert not tmp_project.manifest_path.is_file()
    assert tmp_project._manifest.contract_types is None
    assert tmp_project.sources._path_cache is None


def test_unpack(project_with_source_files_contract):
    with create_tempdir() as path:
        project_with_source_files_contract.unpack(path)
        assert (path / "contracts" / "Contract.json").is_file()

        # Show that even non-sources end up in the unpacked destination.
        assert (path / "contracts" / "Path.with.sub.json").is_file()


def test_unpack_includes_build_file(project_with_contracts):
    with project_with_contracts.isolate_in_tempdir() as tmp_project:
        buildfile = tmp_project.path / ".build" / "__local__.json"
        buildfile.parent.mkdir(parents=True, exist_ok=True)
        buildfile.write_text("{}", encoding="utf8")

        with create_tempdir() as path:
            tmp_project.unpack(path)
            assert (path / ".build" / "__local__.json").is_file()


def test_unpack_includes_interfaces():
    iname = "unp_interface"
    with create_tempdir() as path:
        interfaces = path / "interfaces"
        interfaces.mkdir(parents=True, exist_ok=True)
        interface = interfaces / f"{iname}.json"
        interface.write_text("{}", encoding="utf8")

        project = Project(path)
        with create_tempdir() as new_path:
            project.unpack(new_path)
            expected_interface = new_path / "interfaces" / f"{iname}.json"
            assert expected_interface.is_file()


def test_unpack_includes_interfaces_when_part_of_contracts():
    iname = "unp_interface"
    with create_tempdir() as path:
        interfaces = path / "contracts" / "interfaces"
        interfaces.mkdir(parents=True, exist_ok=True)
        interface = interfaces / f"{iname}.json"
        interface.write_text("{}", encoding="utf8")

        project = Project(path)
        with create_tempdir() as new_path:
            project.unpack(new_path)
            expected_interface = new_path / "contracts" / "interfaces" / f"{iname}.json"
            assert expected_interface.is_file()


def test_add_compiler_data(project_with_dependency_config):
    # NOTE: Using different project than default to lessen
    #   chance of race-conditions from multiprocess test runners.
    project = project_with_dependency_config

    # Load contracts so that any compilers that may exist are present.
    project.load_contracts()
    start_compilers = project.manifest.compilers or []

    # NOTE: Pre-defining things to lessen chance of race condition.
    compiler = Compiler(
        name="comp",
        version="1.0.0",
        contractTypes=["foo"],
        settings={"outputSelection": {"path/to/Foo.sol": "*"}},
    )
    compiler_2 = Compiler(
        name="test",
        version="2.0.0",
        contractTypes=["bar", "stay"],
        settings={"outputSelection": {"path/to/Bar.vy": "*", "stay.vy": "*"}},
    )

    # NOTE: Has same contract as compiler 2 and thus replaces the contract.
    compiler_3 = Compiler(
        name="test",
        version="3.0.0",
        contractTypes=["bar"],
        settings={"outputSelection": {"path/to/Bar.vy": "*"}},
    )

    argument = [compiler]
    second_arg = [compiler_2]
    third_arg = [compiler_3]
    first_exp = [*start_compilers, compiler]
    final_exp = [*first_exp, compiler_2]

    # Ensure types are in manifest for type-source-id lookup.
    bar = ContractType(contractName="bar", sourceId="path/to/Bar.vy")
    foo = ContractType(contractName="foo", sourceId="path/to/Foo.sol")
    project._manifest = PackageManifest(
        contractTypes={"bar": bar, "foo": foo},
        sources={"path/to/Bar.vy": Source(), "path/to/Foo.vy": Source()},
    )
    project._contracts = project._manifest.contract_types
    assert project._manifest.contract_types, "Setup failed - need manifest contract types"

    # Add twice to show it's only added once.
    project.add_compiler_data(argument)
    project.add_compiler_data(argument)
    assert project.manifest.compilers == first_exp

    # NOTE: `add_compiler_data()` will not override existing compilers.
    #   Use `update_cache()` for that.
    project.add_compiler_data(second_arg)
    assert project.manifest.compilers == final_exp
    project.add_compiler_data(third_arg)
    comp = [c for c in project.manifest.compilers if c.name == "test" and c.version == "2.0.0"][0]
    assert "bar" not in comp.contractTypes
    assert "path/to/Bar.vy" not in comp.settings["outputSelection"]
    new_comp = [c for c in project.manifest.compilers if c.name == "test" and c.version == "3.0.0"][
        0
    ]
    assert "bar" in new_comp.contractTypes
    assert "path/to/Bar.vy" in new_comp.settings["outputSelection"]

    # Show that compilers without contract types go away.
    (compiler_3.contractTypes or []).append("stay")
    project.add_compiler_data(third_arg)
    comp_check = [
        c for c in project.manifest.compilers if c.name == "test" and c.version == "2.0.0"
    ]
    assert not comp_check

    # Show error on multiple of same compiler.
    compiler_4 = Compiler(name="test123", version="3.0.0", contractTypes=["bar"])
    compiler_5 = Compiler(name="test123", version="3.0.0", contractTypes=["baz"])
    with pytest.raises(ProjectError, match=r".*was given multiple of the same compiler.*"):
        project.add_compiler_data([compiler_4, compiler_5])

    # Show error when contract type collision (only happens with inputs, else latter replaces).
    compiler_4 = Compiler(name="test321", version="3.0.0", contractTypes=["bar"])
    compiler_5 = Compiler(name="test456", version="9.0.0", contractTypes=["bar"])
    with pytest.raises(ProjectError, match=r".*'bar' collision across compilers.*"):
        project.add_compiler_data([compiler_4, compiler_5])


def test_project_api_foundry_and_ape_config_found(foundry_toml):
    """
    If a project is both a Foundry project and an Ape project,
    ensure both configs are honored.
    """
    with create_tempdir() as tmpdir:
        foundry_cfg_file = tmpdir / "foundry.toml"
        foundry_cfg_file.write_text(foundry_toml, encoding="utf8")

        ape_cfg_file = tmpdir / "ape-config.yaml"
        ape_cfg_file.write_text("name: testfootestfootestfoo", encoding="utf8")

        temp_project = Project(tmpdir)
        assert isinstance(temp_project.project_api, MultiProject)

        # This came from the ape config file.
        assert temp_project.config.name == "testfootestfootestfoo"

        # This came from the foundry toml file.
        assert temp_project.config.contracts_folder == "src"
        assert len(temp_project.config.solidity.import_remapping) > 0


def test_get_contract(project_with_contracts):
    actual = project_with_contracts.get_contract("Other")
    assert isinstance(actual, ContractContainer), f"{type(actual)}"
    assert actual.contract_type.name == "Other"

    # Ensure manifest is only loaded once by none-ing out the path.
    # Otherwise, this can be a MAJOR performance hit.
    manifest_path = project_with_contracts.manifest_path
    project_with_contracts.manifest_path = None
    try:
        actual = project_with_contracts.get_contract("Other")
        assert isinstance(actual, ContractContainer)
        assert actual.contract_type.name == "Other"
    finally:
        project_with_contracts.manifest_path = manifest_path


def test_get_contract_not_exists(project):
    actual = project.get_contract("this is not a contract")
    assert actual is None


class TestProject:
    """
    All tests related to ``ape.Project``.
    """

    def test_init(self, with_dependencies_project_path):
        # Purpose not using `project_with_contracts` fixture.
        project = Project(with_dependencies_project_path)

        # NOTE: Using tempdir to avoid clashing with other tests during x-dist.
        with project.isolate_in_tempdir() as temp_project:
            assert project.path == with_dependencies_project_path
            project.manifest_path.unlink(missing_ok=True)

            #  Re-init to show it doesn't create the manifest file.
            project = Project(temp_project.path)

    def test_init_invalid_config(self):
        here = os.curdir
        with create_tempdir() as temp_dir:
            cfgfile = temp_dir / "ape-config.yaml"
            # Name is invalid!
            cfgfile.write_text("name:\n  {asdf}")

            os.chdir(temp_dir)
            expected = r"[.\n]*Input should be a valid string\n-->1: name:\n   2:   {asdf}[.\n]*"
            weird_project = Project(temp_dir)
            try:
                with pytest.raises(ConfigError, match=expected):
                    _ = weird_project.path
            finally:
                os.chdir(here)

    def test_config_override(self, with_dependencies_project_path):
        contracts_folder = with_dependencies_project_path / "my_contracts"
        config = {"contracts_folder": contracts_folder.name}
        project = Project(with_dependencies_project_path, config_override=config)
        assert project.contracts_folder == contracts_folder

    def test_from_manifest(self, manifest):
        # Purposely not using `project_from_manifest` fixture.
        project = Project.from_manifest(manifest)
        assert isinstance(project, Project)
        assert project.manifest == manifest

    def test_from_manifest_contracts_iter(self, contract_type, project_from_manifest):
        actual = set(iter(project_from_manifest.contracts))
        assert actual == {"FooContractFromManifest"}

    def test_from_manifest_getattr(self, contract_type, project_from_manifest):
        expected = ContractContainer(contract_type)
        actual = project_from_manifest.FooContractFromManifest
        assert isinstance(actual, ContractContainer)
        assert actual == expected

    def test_from_manifest_getitem(self, contract_type, project_from_manifest):
        expected = ContractContainer(contract_type)
        assert project_from_manifest["FooContractFromManifest"] == expected

    def test_from_manifest_load_contracts(self, contract_type):
        """
        Show if contract-types are missing but sources set,
        compiling will add contract-types.
        """
        manifest = make_manifest(contract_type, include_contract_type=False)
        project = Project.from_manifest(manifest)
        assert not project.manifest.contract_types, "Setup failed"

        # Returns containers, not types.
        actual = project.load_contracts()
        assert actual[contract_type.name].contract_type == contract_type

        # Also, show it got set on the manifest.
        assert project.manifest.contract_types == {contract_type.name: contract_type}

    def test_from_python_library(self):
        # web3py as an ape-project dependency.
        web3 = Project.from_python_library("web3")
        assert "site-packages" in str(web3.path)
        assert web3.path.name == "web3"

    def test_hash(self, with_dependencies_project_path, project_from_manifest):
        """
        Show we can use projects in sets.
        """
        project_0 = Project(with_dependencies_project_path)
        project_1 = Project.from_python_library("web3")
        project_2 = project_from_manifest

        # Show we can use in sets.
        project_set = {project_0, project_1, project_2}
        assert len(project_set) == 3

        # Show we can use as dict-keys:
        project_dict = {project_0: 123}
        assert project_dict[project_0] == 123


class TestBrownieProject:
    """
    Tests related to the brownie implementation of the ProjectAPI.
    """

    @pytest.fixture
    def brownie_project(self, base_projects_directory):
        project_path = base_projects_directory / "BrownieProject"
        return BrownieProject(path=project_path)

    def test_extract_config(self, config, brownie_project):
        config = brownie_project.extract_config()

        # Ensure contracts_folder works.
        assert config.contracts_folder == "contractsrenamed"

        # Ensure Solidity and dependencies configuration mapped correctly
        assert config.solidity.version == "0.6.12"

        # NOTE: `contracts/` is not part of the import key as it is
        # usually included in the import statements.
        assert [str(x) for x in config.solidity.import_remapping] == [
            "@openzeppelin=openzeppelin/3.1.0"
        ]
        assert config.dependencies[0]["name"] == "openzeppelin"
        assert config.dependencies[0]["github"] == "OpenZeppelin/openzeppelin-contracts"
        assert config.dependencies[0]["version"] == "3.1.0"


class TestFoundryProject:
    @pytest.fixture
    def mock_github(self, mocker):
        return mocker.MagicMock()

    @pytest.fixture(scope="class")
    def gitmodules(self):
        return """
[submodule "lib/forge-std"]
    path = lib/forge-std
    url = https://github.com/foundry-rs/forge-std
    branch = v1.5.2
[submodule "lib/openzeppelin-contracts"]
    path = lib/openzeppelin-contracts
    url = https://github.com/OpenZeppelin/openzeppelin-contracts
    release = v4.9.5
    branch = v4.9.5
[submodule "lib/erc4626-tests"]
    path = lib/erc4626-tests
    url = https://github.com/a16z/erc4626-tests.git
""".lstrip().replace("    ", "\t")

    def test_extract_config(self, foundry_toml, gitmodules, mock_github):
        with create_tempdir() as temp_dir:
            cfg_file = temp_dir / "foundry.toml"
            cfg_file.write_text(foundry_toml, encoding="utf8")
            gitmodules_file = temp_dir / ".gitmodules"
            gitmodules_file.write_text(gitmodules, encoding="utf8")
            temp_project = Project(temp_dir)

            api = temp_project.project_api
            mock_github.get_repo.return_value = {"default_branch": "main"}
            api._github_client = mock_github  # type: ignore
            assert isinstance(api, FoundryProject)

            # Ensure solidity config migrated.
            actual = temp_project.config.model_dump(
                by_alias=True
            )  # Is result of ``api.extract_config()``.
            assert actual["contracts_folder"] == "src"
            assert "solidity" in actual, "Solidity failed to migrate"
            actual_sol = actual["solidity"]
            assert actual_sol["import_remapping"] == [
                "@openzeppelin=src/.cache/openzeppelin/v4.9.5/",
                "forge-std=src/.cache/forge-std/v1.5.2/src",
            ]
            assert actual_sol["version"] == "0.8.18"
            assert actual_sol["evm_version"] == "cancun"
            assert actual_sol["via_ir"] is True

            # Ensure dependencies migrated from .gitmodules.
            assert "dependencies" in actual, "Dependencies failed to migrate"
            actual_dependencies = actual["dependencies"]
            expected_dependencies = [
                {"github": "foundry-rs/forge-std", "name": "forge-std", "ref": "v1.5.2"},
                {
                    "github": "OpenZeppelin/openzeppelin-contracts",
                    "name": "openzeppelin",
                    "version": "v4.9.5",
                },
                {"github": "a16z/erc4626-tests", "name": "erc4626-tests", "ref": "main"},
            ]
            assert actual_dependencies == expected_dependencies


class TestSourceManager:
    def test_lookup(self, tmp_project):
        source_id = tmp_project.Other.contract_type.source_id
        path = tmp_project.sources.lookup(source_id)
        assert path == tmp_project.path / source_id

    def test_lookup_same_source_id_as_local_project(self, project, tmp_project):
        """
        Tests against a bug where if the source ID of the project matched
        a file in the local project, it would mistakenly return the path
        to the local file instead of the project's file.
        """
        source_id = project.contracts["ContractA"].source_id
        path = project.sources.lookup(source_id)
        assert path.is_file(), "Test path does not exist."

        cfg = {"contracts_folder": project.config.contracts_folder}
        with Project.create_temporary_project(config_override=cfg) as other_tmp_project:
            new_source = other_tmp_project.path / source_id
            new_source.parent.mkdir(parents=True, exist_ok=True)
            new_source.write_text(path.read_text(encoding="utf8"), encoding="utf8")

            actual = other_tmp_project.sources.lookup(source_id)
            assert actual is not None
            expected = other_tmp_project.path / source_id
            assert actual == expected

    def test_lookup_missing_extension(self, tmp_project):
        source_id = tmp_project.Other.contract_type.source_id
        source_id_wo_ext = ".".join(source_id.split(".")[:-1])
        path = tmp_project.sources.lookup(source_id_wo_ext)
        assert path == tmp_project.path / source_id

    def test_lookup_mismatched_extension(self, tmp_project):
        source_id = tmp_project.Other.contract_type.source_id
        source_id = source_id.replace(".json", ".js")
        path = tmp_project.sources.lookup(source_id)
        assert path is None

    def test_lookup_closest_match(self, project_with_source_files_contract):
        pm = project_with_source_files_contract
        source_path = pm.contracts_folder / "Contract.json"
        temp_dir_a = pm.contracts_folder / "temp"
        temp_dir_b = temp_dir_a / "tempb"
        nested_source_a = temp_dir_a / "Contract.json"
        nested_source_b = temp_dir_b / "Contract.json"

        def clean():
            # NOTE: Will also delete temp_dir_b.
            if temp_dir_a.is_dir():
                shutil.rmtree(temp_dir_a)

        clean()

        # NOTE: Will also make temp_dir_a.
        temp_dir_b.mkdir(parents=True)

        try:
            # Duplicate contract so that there are multiple with the same name.
            for nested_src in (nested_source_a, nested_source_b):
                nested_src.touch()
                nested_src.write_text(source_path.read_text(), encoding="utf8")

            # Top-level match.
            for base in (source_path, str(source_path), "Contract", "Contract.json"):
                # Using stem in case it returns `Contract.__mock__`, which is
                # added / removed as part of other tests (running x-dist).
                assert pm.sources.lookup(base).stem == source_path.stem, f"Failed to lookup {base}"

            # Nested: 1st level
            for closest in (
                nested_source_a,
                str(nested_source_a),
                "temp/Contract",
                "temp/Contract.json",
            ):
                actual = pm.sources.lookup(closest)
                expected = nested_source_a
                # Using stem in case it returns `Contract.__mock__`, which is
                # added / removed as part of other tests (running x-dist).
                assert actual.stem == expected.stem, f"Failed to lookup {closest}"

            # Nested: 2nd level
            for closest in (
                nested_source_b,
                str(nested_source_b),
                "temp/tempb/Contract",
                "temp/tempb/Contract.json",
            ):
                actual = pm.sources.lookup(closest)
                expected = nested_source_b

                # Using stem in case it returns `Contract.__mock__`, which is
                # added / removed as part of other tests (running x-dist).
                assert actual.stem == expected.stem, f"Failed to lookup {closest}"

        finally:
            clean()

    def test_lookup_not_found(self, tmp_project):
        assert tmp_project.sources.lookup("madeup.json") is None

    def test_lookup_missing_contracts_prefix(self, project_with_source_files_contract):
        """
        Show we can exclude the `contracts/` prefix in a source ID.
        """
        project = project_with_source_files_contract
        actual_from_str = project.sources.lookup("ContractA.sol")
        actual_from_path = project.sources.lookup(Path("ContractA.sol"))
        expected = project.contracts_folder / "ContractA.sol"
        assert actual_from_str == actual_from_path == expected
        assert actual_from_str.is_absolute()
        assert actual_from_path.is_absolute()

    def test_paths_exclude(self, tmp_project):
        exclude_config = {"compile": {"exclude": ["Other.json"]}}
        with tmp_project.temp_config(**exclude_config):
            # Show default excludes also work, such as a .DS_Store file.
            ds_store = tmp_project.contracts_folder / ".DS_Store"
            ds_store.write_bytes(b"asdfasf")

            # Show anything in compiler-cache is ignored.
            cache = tmp_project.contracts_folder / ".cache"
            cache.mkdir(exist_ok=True)
            random_file = cache / "dontmindme.json"
            random_file.write_text("what, this isn't json?!", encoding="utf8")

            path_ids = {
                f"{tmp_project.contracts_folder.name}/{src.name}"
                for src in tmp_project.sources.paths
            }
            excluded = {".DS_Store", "Other.json", ".cache/dontmindme.json"}
            for actual in (path_ids, tmp_project.sources):
                for exclusion in excluded:
                    expected = f"{tmp_project.contracts_folder.name}/{exclusion}"
                    assert expected not in actual

    def test_is_excluded(self, project_with_contracts):
        exclude_cfg = {"compile": {"exclude": ["exclude_dir/*", "Excl*.json"]}}
        source_ids = ("contracts/exclude_dir/UnwantedContract.json", "contracts/Exclude.json")
        with project_with_contracts.temp_config(**exclude_cfg):
            for source_id in source_ids:
                path = project_with_contracts.path / source_id
                assert project_with_contracts.sources.is_excluded(path)

    def test_items(self, project_with_contracts):
        actual = list(project_with_contracts.sources.items())
        assert len(actual) > 0
        assert isinstance(actual[0], tuple)
        assert "contracts/Other.json" in [x[0] for x in actual]
        assert isinstance(actual[0][1], Source)

    def test_keys(self, project_with_contracts):
        actual = list(project_with_contracts.sources.keys())
        assert "contracts/Other.json" in actual

    def test_values(self, project_with_contracts):
        actual = list(project_with_contracts.sources.values())
        assert all(isinstance(x, Source) for x in actual)


class TestContractManager:
    def test_iter(self, tmp_project):
        actual = set(iter(tmp_project.contracts))
        assert actual == {"Project", "Other"}

    def test_compile(self, tmp_project):
        actual = list(tmp_project.contracts._compile("contracts/Project.json"))
        assert len(actual) == 1
        assert actual[0].contract_type.name == "Project"

        # Show it can happen again.
        actual = list(tmp_project.contracts._compile("contracts/Project.json"))
        assert len(actual) == 1
        assert actual[0].contract_type.name == "Project"

    def test_values(self, tmp_project):
        contracts = [c for c in tmp_project.contracts.values()]
        actual = {x.name for x in contracts}
        assert len(actual) == 2
        assert actual == {"Other", "Project"}
        # Delete a file and try again, as a test.
        file = tmp_project.path / contracts[0].source_id
        file.unlink()

        contracts = [c for c in tmp_project.contracts.values()]
        actual = {x.name for x in contracts}
        assert len(actual) == 1
        assert file.name not in actual


class TestDeploymentManager:
    @pytest.fixture
    def project(self, tmp_project, vyper_contract_instance):
        contract_type = vyper_contract_instance.contract_type
        tmp_project.manifest.contract_types = {contract_type.name: contract_type}
        return tmp_project

    def test_track(self, project, vyper_contract_instance):
        project.deployments.track(vyper_contract_instance, allow_dev=True)
        deployment = next(iter(project.deployments), None)
        contract_type = vyper_contract_instance.contract_type
        assert deployment is not None
        assert deployment.contract_type == f"{contract_type.source_id}:{contract_type.name}"

    def test_instance_map(self, project, vyper_contract_instance):
        project.deployments.track(vyper_contract_instance, allow_dev=True)
        assert project.deployments.instance_map != {}

        bip122_chain_id = to_hex(project.provider.get_block(0).hash)
        expected_uri = f"blockchain://{bip122_chain_id[2:]}/block/"
        for key in project.deployments.instance_map.keys():
            if key.startswith(expected_uri):
                return

        assert False, "Failed to find expected URI"


def test_chdir(project):
    original_path = project.path
    with create_tempdir() as new_path:
        project.chdir(new_path)
        assert project.path == new_path

        # Undo.
        project.chdir(original_path)
        assert project.path == original_path

        # Show you can also use it in a context.
        with project.chdir(new_path):
            assert project.path == new_path

        # It should have automatically undone.
        assert project.path == original_path


def test_within_project_path():
    start_cwd = Path.cwd()
    with create_tempdir() as new_path:
        project = Project(new_path)
        assert Path.cwd() != new_path

        with project.within_project_path():
            assert Path.cwd() == project.path

    assert Path.cwd() == start_cwd



import os
import signal
from collections.abc import Collection, Iterator
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from evmchains import PUBLIC_CHAIN_META
from pydantic import ValidationError

from ape.api.networks import ProviderContextManager
from ape.exceptions import EcosystemNotFoundError, NetworkError, NetworkNotFoundError
from ape.managers.base import BaseManager
from ape.utils.basemodel import (
    BaseModel,
    DiskCacheableModel,
    ExtraAttributesMixin,
    ExtraModelAttributes,
    get_attribute_with_extras,
    only_raise_attribute_error,
)
from ape.utils.misc import _dict_overlay, log_instead_of_fail
from ape.utils.os import clean_path
from ape_ethereum.provider import EthereumNodeProvider

if TYPE_CHECKING:
    from ape.api.networks import EcosystemAPI, NetworkAPI
    from ape.api.providers import ProviderAPI, SubprocessProvider
    from ape.utils.rpc import RPCHeaders


class NodeProcessData(BaseModel):
    """
    Cached data for node subprocesses managed by Ape.
    """

    network_choice: str
    """
    The network triple ``ecosystem:network:node``.
    """

    ipc_path: Optional[Path] = None
    """
    The IPC path this node process communicates on.
    """

    http_uri: Optional[str] = None
    """
    The HTTP URI this node process exposes.
    """

    ws_uri: Optional[str] = None
    """
    The websockets URI this node process exposes.
    """

    @log_instead_of_fail(default="<NodeProcessData>")
    def __repr__(self) -> str:
        if ipc := self.ipc_path:
            return f"{self.network_choice} - {clean_path(ipc)}"

        elif uri := (self.http_uri or self.ws_uri):
            return f"{self.network_choice} - {uri}"

        return self.network_choice

    def matches_provider(self, provider: "SubprocessProvider") -> bool:
        if self.network_choice != f"{provider.network.choice}:{provider.name}":
            return False

        # Skip if any of the connection paths (IPC, HTTP, WS) differ
        for attr in ("ipc_path", "http_uri", "ws_uri"):
            if getattr(provider, attr, None) != getattr(self, attr, None):
                return False

        return True


class NodeProcessMap(DiskCacheableModel):
    """
    All managed running network subprocesses.
    """

    nodes: dict[int, NodeProcessData] = {}

    @property
    def path(self) -> Path:
        return self._path

    @property
    def process_ids(self) -> tuple[int, ...]:
        return tuple(self.nodes.keys())

    def __bool__(self) -> bool:
        """
        True if there is at least one managed node process.
        """
        return bool(self.nodes)

    def __contains__(self, pid_or_provider: Union[int, "SubprocessProvider"]) -> bool:
        if isinstance(pid_or_provider, int):
            return pid_or_provider in self.nodes

        for data in self.nodes.values():
            if data.matches_provider(pid_or_provider):
                return True

        return False

    def get(self, pid: Union[int, str]) -> Optional[NodeProcessData]:
        return self.nodes.get(int(pid))

    def lookup_processes(self, provider: "SubprocessProvider") -> dict[int, NodeProcessData]:
        return {pid: data for pid, data in self.nodes.items() if data.matches_provider(provider)}

    def cache_provider(self, provider: "SubprocessProvider"):
        # Don't use `provider.network_choice` here because we want to ensure the provider
        # name is used and not the URI for the third part of the choice.
        network_choice = f"{provider.network.choice}:{provider.name}"
        data: dict[str, Any] = {
            "network_choice": network_choice,
            "ipc_path": provider.ipc_path,
            "http_uri": provider.http_uri,
            "ws_uri": provider.ws_uri,
        }
        data = {k: v for k, v in data.items() if v is not None}

        if process := provider.process:
            obj = NodeProcessData.model_validate(data)

            # Ensure no duplicates (bad clean?)
            self._delete_all_matching(obj)

            self.nodes[process.pid] = obj
            self.model_dump_file()

        else:
            raise NetworkError("Unable to cache subprocess-provider information: not connected.")

    def _delete_all_matching(self, node: NodeProcessData):
        pids_to_remove = set()
        for pid, other_node in self.nodes.items():
            if other_node == node:
                # Exact match but different PID. Bad clean.
                pids_to_remove.add(pid)

        for pid in pids_to_remove:
            del self.nodes[pid]

    def remove_provider(self, provider: "SubprocessProvider"):
        pids_to_remove = {
            pid for pid, data in self.nodes.items() if data.matches_provider(provider)
        }
        self.nodes = {pid: node for pid, node in self.nodes.items() if pid not in pids_to_remove}
        self.model_dump_file()

    def remove_processes(self, *pids: int):
        self.nodes = {k: v for k, v in self.nodes.items() if k not in pids}
        self.model_dump_file()

    def clean(self):
        self._path.unlink(missing_ok=True)


class NetworkManager(BaseManager, ExtraAttributesMixin):
    """
    The set of all blockchain network ecosystems registered from the plugin system.
    Typically, you set the provider via the ``--network`` command line option.
    However, use this singleton for more granular access to networks.

    Usage example::

        from ape import networks

        # "networks" is the NetworkManager singleton
        with networks.ethereum.mainnet.use_provider("node"):
            ...
    """

    _active_provider: Optional["ProviderAPI"] = None
    _default_ecosystem_name: Optional[str] = None

    # For adhoc adding custom networks, or incorporating some defined
    # in other projects' configs.
    _custom_networks: list[dict] = []

    @log_instead_of_fail(default="<NetworkManager>")
    def __repr__(self) -> str:
        provider = self.active_provider
        class_name = NetworkManager.__name__
        content = f"{class_name} active_provider={provider!r}" if provider else class_name
        return f"<{content}>"

    @property
    def active_provider(self) -> Optional["ProviderAPI"]:
        """
        The currently connected provider if one exists. Otherwise, returns ``None``.
        """
        return self._active_provider

    @active_provider.setter
    def active_provider(self, new_value: "ProviderAPI"):
        self._active_provider = new_value

    @property
    def connected(self) -> bool:
        """
        ``True`` when there is an active provider.
        """
        return self.active_provider is not None

    @property
    def network(self) -> "NetworkAPI":
        """
        The current network if connected to one.

        Raises:
            :class:`~ape.exceptions.ProviderNotConnectedError`: When there is
              no active provider at runtime.

        Returns:
            :class:`~ape.api.networks.NetworkAPI`
        """
        return self.provider.network

    @property
    def ecosystem(self) -> "EcosystemAPI":
        """
        The current ecosystem if connected to one.

        Raises:
            :class:`~ape.exceptions.ProviderNotConnectedError`: When there is
              no active provider at runtime.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`
        """
        return self.network.ecosystem

    @cached_property
    def running_nodes(self) -> NodeProcessMap:
        """
        All running development nodes managed by Ape.
        """
        path = self.config_manager.DATA_FOLDER / "processes" / "nodes.json"
        try:
            return NodeProcessMap.model_validate_file(path)
        except ValidationError:
            path.unlink(missing_ok=True)
            return NodeProcessMap.model_validate_file(path)

    def get_running_node(self, pid: int) -> "SubprocessProvider":
        """
        Get a running subprocess provider for the given ``pid``.

        Args:
            pid (int): The process ID.

        Returns:
            class:`~ape.api.providers.SubprocessProvider`
        """
        if not (data := self.running_nodes.get(pid)):
            raise NetworkError(f"No running node for pid '{pid}'.")

        uri: Optional[Union[str, Path]] = None
        if ipc := data.ipc_path:
            if ipc.exists():
                uri = ipc

        else:
            uri = data.http_uri or data.ws_uri

        if uri is None:
            NetworkError(f"Cannot connect to node on PID '{pid}': Missing URI data.")

        # In this case, we want the more connectable network choice.
        network_parts = data.network_choice.split(":")
        network_choice = f"{':'.join(network_parts[:2])}:{uri}"

        provider_settings: dict = {
            network_parts[0]: {
                network_parts[1]: {
                    "ipc_path": data.ipc_path,
                    "http_uri": data.http_uri,
                    "ws_uri": data.ws_uri,
                    "uri": None,
                }
            }
        }
        provider = self.get_provider_from_choice(
            network_choice=network_choice, provider_settings=provider_settings or None
        )

        # If this is not a subprocess provider, it may be ok to proceed.
        # However, the rest of Ape will assume it is.
        return provider  # type: ignore[return-value]

    def kill_node_process(self, *process_ids: int) -> dict[int, NodeProcessData]:
        """
        Kill a node process managed by Ape.

        Args:
            *process_ids (int): The process ID to kill.

        Returns:
            dict[str, :class:`~ape.managers.networks.NodeProcessData`]: The process data
            of all terminated processes.
        """
        if not self.running_nodes:
            return {}

        pids_killed = {}
        for pid in process_ids:
            if not (data := self.running_nodes.nodes.get(pid)):
                continue

            try:
                provider = self.get_running_node(pid)
            except Exception:
                # Still try to kill the process (below).
                pass
            else:
                # Gracefully disconnect _before_ killing process.
                provider.disconnect()

            try:
                os.kill(pid, signal.SIGTERM)
            except Exception:
                pass
            else:
                pids_killed[pid] = data

        self.running_nodes.remove_processes(*process_ids)
        return pids_killed

    def get_request_headers(
        self, ecosystem_name: str, network_name: str, provider_name: str
    ) -> "RPCHeaders":
        """
        All request headers to be used when connecting to this network.
        """
        ecosystem = self.get_ecosystem(ecosystem_name)
        network = ecosystem.get_network(network_name)
        provider = network.get_provider(provider_name)
        headers = self.config_manager._get_request_headers()
        for obj in (ecosystem, network, provider):
            for key, value in obj._get_request_headers().items():
                headers[key] = value

        return headers

    def fork(
        self,
        provider_name: Optional[str] = None,
        provider_settings: Optional[dict] = None,
        block_number: Optional[int] = None,
    ) -> ProviderContextManager:
        """
        Fork the currently connected network.

        Args:
            provider_name (str, optional): The name of the provider to get. Defaults to ``None``.
              When ``None``, returns the default provider.
            provider_settings (dict, optional): Settings to apply to the provider. Defaults to
              ``None``.
            block_number (Optional[int]): Optionally specify the block number you wish to fork.
              Negative block numbers are relative to HEAD. Defaults to the configured fork
              block number or HEAD.

        Returns:
            :class:`~ape.api.networks.ProviderContextManager`
        """
        network_name = self.network.name
        is_fork_already = network_name.endswith("-fork")
        forked_network_name = network_name if is_fork_already else f"{network_name}-fork"
        try:
            forked_network = self.ecosystem.get_network(forked_network_name)
        except NetworkNotFoundError as err:
            raise NetworkError(f"Unable to fork network '{network_name}'.") from err

        provider_settings = provider_settings or {}
        if is_fork_already and "host" not in provider_settings:
            # Forking a fork- to ensure is using a different Port,
            # use the "auto-port" feature.
            provider_settings["host"] = "auto"

        fork_settings = {}
        if block_number is not None:
            # Negative block_number means relative to HEAD
            if block_number < 0:
                latest_block_number = self.provider.get_block("latest").number or 0
                block_number = latest_block_number + block_number
                if block_number < 0:
                    # If the block number is still negative, they have forked past genesis.
                    raise NetworkError("Unable to fork past genesis block.")

            # Ensure block_number is set in config for this network
            fork_settings["block_number"] = block_number

        if uri := self.provider.connection_str:
            fork_settings["upstream_provider"] = uri

        _dict_overlay(
            provider_settings,
            {"fork": {self.ecosystem.name: {self.network.name: fork_settings}}},
        )

        shared_kwargs: dict = {
            "provider_settings": provider_settings,
            "disconnect_after": True,
        }
        return (
            forked_network.use_provider(provider_name, **shared_kwargs)
            if provider_name
            else forked_network.use_default_provider(**shared_kwargs)
        )

    @property
    def ecosystem_names(self) -> set[str]:
        """
        The set of all ecosystem names in ``ape``.
        """
        return set(self.ecosystems)

    @property
    def network_names(self) -> set[str]:
        """
        The set of all network names in ``ape``.
        """

        return {n for e in self.ecosystems.values() for n in e.networks}

    @property
    def provider_names(self) -> set[str]:
        """
        The set of all provider names in ``ape``.
        """

        return set(
            provider
            for ecosystem in self.ecosystems.values()
            for network in ecosystem.networks.values()
            for provider in network.providers
        )

    @property
    def custom_networks(self) -> list[dict]:
        """
        Custom network data defined in various ape-config files
        or added adhoc to the network manager.
        """
        return [*self._custom_networks_from_config, *self._custom_networks]

    @cached_property
    def _custom_networks_from_config(self) -> list[dict]:
        return [
            n.model_dump(by_alias=True)
            for n in self.config_manager.get_config("networks").get("custom", [])
        ]

    @property
    def ecosystems(self) -> dict[str, "EcosystemAPI"]:
        """
        All the registered ecosystems in ``ape``, such as ``ethereum``.
        """
        return {
            **self._evmchains_ecosystems,
            **self._plugin_ecosystems,
            **self._custom_ecosystems,
        }

    @cached_property
    def _plugin_ecosystems(self) -> dict[str, "EcosystemAPI"]:
        # Load plugins (possibly for first time).
        plugins: list[tuple] = self.plugin_manager.ecosystems
        return {n: cls(name=n) for n, cls in plugins}

    @cached_property
    def _custom_ecosystems(self) -> dict[str, "EcosystemAPI"]:
        custom_networks: list = self.custom_networks
        plugin_ecosystems = self._plugin_ecosystems
        evm_chains = self._evmchains_ecosystems
        custom_ecosystems: dict[str, EcosystemAPI] = {}
        for custom_network in custom_networks:
            ecosystem_name = custom_network["ecosystem"]
            if (
                ecosystem_name in plugin_ecosystems
                or ecosystem_name in evm_chains
                or ecosystem_name in custom_ecosystems
            ):
                # Already included in a prior network.
                continue

            base_ecosystem_name = (
                custom_network.get("base_ecosystem_plugin") or self.default_ecosystem_name
            )

            if base_ecosystem_name not in plugin_ecosystems:
                name = custom_network.get("name", "?")
                if eco := custom_network.get("ecosystem"):
                    name = f"{eco}:{name}"

                msg = (
                    f"Custom network '{name}' specified unknown base-ecosystem class "
                    f"'{base_ecosystem_name}'. Are you missing plugin 'ape-{base_ecosystem_name}'?"
                )
                raise NetworkError(msg)

            existing_cls = plugin_ecosystems[base_ecosystem_name]
            ecosystem_cls = existing_cls.model_copy(
                update={"name": ecosystem_name},
                cache_clear=("_networks_from_plugins", "_networks_from_evmchains"),
            )
            custom_ecosystems[ecosystem_name] = ecosystem_cls

        return custom_ecosystems

    @cached_property
    def _evmchains_ecosystems(self) -> dict[str, "EcosystemAPI"]:
        ecosystems: dict[str, EcosystemAPI] = {}
        for name in PUBLIC_CHAIN_META:
            ecosystem_name = name.lower().replace(" ", "-")
            symbol = None
            for net in PUBLIC_CHAIN_META[ecosystem_name].values():
                if not (native_currency := net.get("nativeCurrency")):
                    continue

                if "symbol" not in native_currency:
                    continue

                symbol = native_currency["symbol"]
                break

            symbol = symbol or "ETH"

            # Is an EVM chain, can automatically make a class using evm-chains.
            evm_class = self._plugin_ecosystems["ethereum"].__class__
            ecosystems[name] = evm_class(name=ecosystem_name, fee_token_symbol=symbol)

        return ecosystems

    def create_custom_provider(
        self,
        connection_str: str,
        provider_cls: type["ProviderAPI"] = EthereumNodeProvider,
        provider_name: Optional[str] = None,
    ) -> "ProviderAPI":
        """
        Create a custom connection to a URI using the EthereumNodeProvider provider.
        **NOTE**: This provider will assume EVM-like behavior and this is generally not recommended.
        Use plugins when possible!

        Args:
            connection_str (str): The connection string of the node, such as its URI
              when using HTTP.
            provider_cls (type[:class:`~ape.api.providers.ProviderAPI`]): Defaults to
              :class:`~ape_ethereum.providers.EthereumNodeProvider`.
            provider_name (Optional[str]): The name of the provider. Defaults to best guess.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`: The Geth provider
              implementation that comes with Ape.
        """

        network = self.ethereum.custom_network

        if provider_name is None:
            if issubclass(provider_cls, EthereumNodeProvider):
                name = "node"

            elif cls_name := getattr(provider_cls, "name", None):
                name = cls_name

            elif cls_name := provider_cls.__name__:
                name = cls_name.lower()

            else:
                # Would be unusual for this to happen though.
                name = "provider"

        else:
            name = provider_name

        provider_settings: dict = {}
        if connection_str.startswith("https://") or connection_str.startswith("http://"):
            provider_settings["uri"] = connection_str
        elif connection_str.endswith(".ipc"):
            provider_settings["ipc_path"] = connection_str
        else:
            raise NetworkError(f"Scheme for '{connection_str}' not yet supported.")

        return (provider_cls or EthereumNodeProvider)(
            name=name,
            network=network,
            provider_settings=provider_settings,
            data_folder=self.ethereum.data_folder / name,
        )

    def __iter__(self) -> Iterator[str]:
        """
        All the managed ecosystems in ``ape``, as an iterable.

        Returns:
            Iterator[:class:`~ape.api.networks.EcosystemAPI`]
        """
        yield from self.ecosystems

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(
            name="ecosystems",
            attributes=lambda: self.ecosystems,
            include_getitem=True,
        )

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> "EcosystemAPI":
        """
        Get an ecosystem via ``.`` access.

        Args:
            attr_name (str): The name of the ecosystem.

        Returns:
            :class:`~ape.api.networks.EcosystemAPI`

        Usage example::

            eth = networks.ethereum
        """
        return get_attribute_with_extras(self, attr_name)

    def get_network_choices(
        self,
        ecosystem_filter: Optional[Union[list[str], str]] = None,
        network_filter: Optional[Union[list[str], str]] = None,
        provider_filter: Optional[Union[list[str], str]] = None,
    ) -> Iterator[str]:
        """
        The set of all possible network choices available as a "network selection"
        e.g. ``--network [ECOSYSTEM:NETWORK:PROVIDER]``.

        Each value is in the form ``ecosystem:network:provider`` and shortened options also
        appear in the list. For example, ``::node`` would default to ``:ethereum:local:node``
        and both will be in the returned list. The values come from each
        :class:`~ape.api.providers.ProviderAPI` that is installed.

        Use the CLI command ``ape networks list`` to list all the possible network
        combinations.

        Args:
            ecosystem_filter (Optional[Union[list[str], str]]): Get only the specified ecosystems.
              Defaults to getting all ecosystems.
            network_filter (Optional[Union[list[str], str]]): Get only the specified networks.
              Defaults to getting all networks in ecosystems.
            provider_filter (Optional[Union[list[str], str]]): Get only the specified providers.
              Defaults to getting all providers in networks.

        Returns:
            Iterator[str]: An iterator over all the network-choice possibilities.
        """

        ecosystem_filter = _validate_filter(ecosystem_filter, self.ecosystem_names)
        network_filter = _validate_filter(network_filter, self.network_names)
        provider_filter = _validate_filter(provider_filter, self.provider_names)

        ecosystem_items = self.ecosystems
        if ecosystem_filter:
            ecosystem_items = {n: e for n, e in ecosystem_items.items() if n in ecosystem_filter}

        for ecosystem_name, ecosystem in ecosystem_items.items():
            network_items = ecosystem.networks
            if network_filter:
                network_items = {n: net for n, net in network_items.items() if n in network_filter}

            if not network_items:
                continue

            ecosystem_has_providers = False
            for network_name, network in network_items.items():
                providers = network.providers
                if provider_filter:
                    providers = [n for n in providers if n in provider_filter]

                network_has_providers = len(providers) > 0
                if not ecosystem_has_providers:
                    # Only check if we still haven't found any
                    ecosystem_has_providers = network_has_providers

                if not network_has_providers:
                    continue

                for provider_name in providers:
                    if (
                        ecosystem_name == self.default_ecosystem.name
                        and network_name == ecosystem.default_network_name
                    ):
                        yield f"::{provider_name}"

                    if ecosystem_name == self.default_ecosystem.name:
                        yield f":{network_name}:{provider_name}"

                    if network_name == ecosystem.default_network_name:
                        yield f"{ecosystem_name}::{provider_name}"

                    # Always include the full path as an option.
                    yield f"{ecosystem_name}:{network_name}:{provider_name}"

                # Providers were yielded if we reached this point.
                if ecosystem_name == self.default_ecosystem.name:
                    yield f":{network_name}"

                yield f"{ecosystem_name}:{network_name}"

            if ecosystem_has_providers:
                yield ecosystem_name

    def get_ecosystem(self, ecosystem_name: str) -> "EcosystemAPI":
        """
        Get the ecosystem for the given name.

        Args:
            ecosystem_name (str): The name of the ecosystem to get.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the ecosystem is not found.

        Returns:
            :class:`~ape.api.networks.EcosystemAPI`
        """
        # NOTE: This method purposely avoids "just checking self.ecosystems"
        #   for performance reasons and exiting the search as early as possible.
        ecosystem_name = ecosystem_name.lower().replace(" ", "-")
        try:
            return self._plugin_ecosystems[ecosystem_name]
        except KeyError:
            pass

        # Check if custom.
        try:
            return self._custom_ecosystems[ecosystem_name]
        except KeyError:
            pass

        if ecosystem := self._get_ecosystem_from_evmchains(ecosystem_name):
            return ecosystem

        raise EcosystemNotFoundError(ecosystem_name, options=self.ecosystem_names)

    def _get_ecosystem_from_evmchains(self, ecosystem_name: str) -> Optional["EcosystemAPI"]:
        if ecosystem_name not in PUBLIC_CHAIN_META:
            return None

        symbol = None
        for net in PUBLIC_CHAIN_META[ecosystem_name].values():
            if not (native_currency := net.get("nativeCurrency")):
                continue

            if "symbol" not in native_currency:
                continue

            symbol = native_currency["symbol"]
            break

        symbol = symbol or "ETH"

        # Is an EVM chain, can automatically make a class using evm-chains.
        evm_class = self._plugin_ecosystems["ethereum"].__class__
        return evm_class(name=ecosystem_name, fee_token_symbol=symbol)

    def get_provider_from_choice(
        self,
        network_choice: Optional[str] = None,
        provider_settings: Optional[dict] = None,
    ) -> "ProviderAPI":
        """
        Get a :class:`~ape.api.providers.ProviderAPI` from a network choice.
        A network choice is any value returned from
        :meth:`~ape.managers.networks.NetworkManager.get_network_choices`. Use the
        CLI command ``ape networks list`` to list all the possible network
        combinations.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given network choice does not
              match any known network.

        Args:
            network_choice (str, optional): The network choice
              (see :meth:`~ape.managers.networks.NetworkManager.get_network_choices`).
              Defaults to the default ecosystem, network, and provider combination.
            provider_settings (dict, optional): Settings for the provider. Defaults to None.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`
        """
        if network_choice is None:
            default_network = self.default_ecosystem.default_network
            return default_network.get_provider(provider_settings=provider_settings)

        elif network_choice.startswith("pid://"):
            # Was given a process ID (already running node on local machine).
            pid_str = network_choice[len("pid://") :]
            if not pid_str.isdigit():
                raise ValueError(f"Invalid PID: {pid_str}")

            return self.get_running_node(int(pid_str))

        elif _is_adhoc_url(network_choice):
            # Custom network w/o ecosystem & network spec.
            return self.create_custom_provider(network_choice)

        selections = network_choice.split(":")

        # NOTE: Handle case when URI is passed e.g. "http://..."
        if len(selections) > 3:
            provider_value = ":".join(selections[2:])
            selections[2] = provider_value
            selections = selections[:3]
            if _is_adhoc_url(provider_value):
                selections[1] = selections[1] or "custom"

        if selections == network_choice or len(selections) == 1:
            # Either split didn't work (in which case it matches the start)
            # or there was nothing after the ``:`` (e.g. "ethereum:")
            ecosystem = self.get_ecosystem(selections[0] or self.default_ecosystem.name)
            # By default, the "local" network should be specified for
            # any ecosystem (this should not correspond to a production chain)
            default_network = ecosystem.default_network
            return default_network.get_provider(provider_settings=provider_settings)

        elif len(selections) == 2:
            # Only ecosystem and network were specified, not provider
            ecosystem_name, network_name = selections
            ecosystem = self.get_ecosystem(ecosystem_name or self.default_ecosystem.name)
            network = ecosystem.get_network(network_name or ecosystem.default_network_name)
            return network.get_provider(provider_settings=provider_settings)

        elif len(selections) == 3:
            # Everything is specified, use specified provider for ecosystem and network
            ecosystem_name, network_name, provider_name = selections
            ecosystem = (
                self.get_ecosystem(ecosystem_name) if ecosystem_name else self.default_ecosystem
            )
            network = ecosystem.get_network(network_name or ecosystem.default_network_name)
            return network.get_provider(
                provider_name=provider_name, provider_settings=provider_settings
            )

        else:
            # NOTE: Might be unreachable
            raise NetworkError("Invalid network selection.")

    def parse_network_choice(
        self,
        network_choice: Optional[str] = None,
        provider_settings: Optional[dict] = None,
        disconnect_after: bool = False,
        disconnect_on_exit: bool = True,
    ) -> ProviderContextManager:
        """
        Parse a network choice into a context manager for managing a temporary
        connection to a provider. See
        :meth:`~ape.managers.networks.NetworkManager.get_network_choices` for all
        available choices (or use CLI command ``ape networks list``).

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given network choice does not
              match any known network.

        Args:
            network_choice (str, optional): The network choice
              (see :meth:`~ape.managers.networks.NetworkManager.get_network_choices`).
              Defaults to the default ecosystem, network, and provider combination.
            provider_settings (dict, optional): Settings for the provider. Defaults to None.
            disconnect_after (bool): Set to True to terminate the connection completely
              at the end of context. NOTE: May only work if the network was also started
              from this session.
            disconnect_on_exit (bool): Whether to disconnect on the exit of the python
              session. Defaults to ``True``.

        Returns:
            :class:`~api.api.networks.ProviderContextManager`
        """
        provider = self.get_provider_from_choice(
            network_choice=network_choice, provider_settings=provider_settings
        )
        return ProviderContextManager(
            provider=provider,
            disconnect_after=disconnect_after,
            disconnect_on_exit=disconnect_on_exit,
        )

    @property
    def default_ecosystem_name(self) -> str:
        if name := self._default_ecosystem_name:
            return name

        return self.local_project.config.default_ecosystem or "ethereum"

    @cached_property
    def default_ecosystem(self) -> "EcosystemAPI":
        """
        The default ecosystem. Call
        :meth:`~ape.managers.networks.NetworkManager.set_default_ecosystem` to
        change the default ecosystem. If a default is not set and there is
        only a single ecosystem installed, such as Ethereum, then get
        that ecosystem.
        """
        return self.get_ecosystem(self.default_ecosystem_name)

    def set_default_ecosystem(self, ecosystem_name: str):
        """
        Change the default ecosystem.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given ecosystem name is unknown.

        Args:
            ecosystem_name (str): The name of the ecosystem to set
              as the default.
        """
        if ecosystem_name in self.ecosystem_names:
            self._default_ecosystem_name = ecosystem_name

        else:
            raise EcosystemNotFoundError(ecosystem_name, options=self.ecosystem_names)

    @property
    def network_data(self) -> dict:
        """
        Get a dictionary containing data about networks in the ecosystem.

        **NOTE**: The keys are added in an opinionated order for nicely
        translating into ``yaml``.

        Returns:
            dict
        """
        return self.get_network_data()

    def get_network_data(
        self,
        ecosystem_filter: Optional[Collection[str]] = None,
        network_filter: Optional[Collection[str]] = None,
        provider_filter: Optional[Collection[str]] = None,
    ):
        data: dict = {"ecosystems": []}

        for ecosystem_name in self:
            if ecosystem_filter and ecosystem_name not in ecosystem_filter:
                continue

            ecosystem_data = self._get_ecosystem_data(
                ecosystem_name,
                network_filter=network_filter,
                provider_filter=provider_filter,
            )
            data["ecosystems"].append(ecosystem_data)

        return data

    def _get_ecosystem_data(
        self,
        ecosystem_name: str,
        network_filter: Optional[Collection[str]] = None,
        provider_filter: Optional[Collection[str]] = None,
    ) -> dict:
        ecosystem = self[ecosystem_name]
        ecosystem_data: dict = {"name": str(ecosystem_name)}

        # Only add isDefault key when True
        if ecosystem_name == self.default_ecosystem.name:
            ecosystem_data["isDefault"] = True

        ecosystem_data["networks"] = []
        networks = getattr(self, ecosystem_name).networks

        for network_name in networks:
            if network_filter and network_name not in network_filter:
                continue

            network_data = ecosystem.get_network_data(network_name, provider_filter=provider_filter)
            ecosystem_data["networks"].append(network_data)

        return ecosystem_data

    def _invalidate_cache(self):
        # NOTE: Called when changing config programmatically.
        self.__dict__.pop("_custom_ecosystems", None)
        self.__dict__.pop("_custom_networks_from_config", None)
        self._custom_networks = []


def _validate_filter(arg: Optional[Union[list[str], str]], options: set[str]):
    filters = arg or []

    if isinstance(filters, str):
        filters = [filters]

    for _filter in filters:
        if _filter not in options:
            raise NetworkError(f"Unknown option '{_filter}'.")

    return filters


def _is_adhoc_url(value: str) -> bool:
    return (
        value.startswith("http://")
        or value.startswith("https://")
        or value.startswith("ws://")
        or value.startswith("wss://")
        or (value.endswith(".ipc") and ":" not in value)
    )



import re
import sys
from collections.abc import Iterable, Iterator, Sequence
from enum import Enum
from functools import cached_property
from shutil import which
from typing import TYPE_CHECKING, Any, Optional
from urllib.parse import urlparse

import click
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from pydantic import field_validator, model_validator

from ape.exceptions import PluginVersionError
from ape.logging import logger
from ape.utils._github import github_client
from ape.utils.basemodel import BaseInterfaceModel, BaseModel
from ape.utils.misc import _get_distributions, get_package_version, log_instead_of_fail
from ape.version import version as ape_version_str

if TYPE_CHECKING:
    from ape.managers.plugins import PluginManager

# Plugins maintained OSS by ApeWorX (and trusted)
# Use `uv pip` if installed, otherwise `python -m pip`
PIP_COMMAND = ["uv", "pip"] if which("uv") else [sys.executable, "-m", "pip"]
PLUGIN_PATTERN = re.compile(r"\bape_\w+(?!\S)")
CORE_PLUGINS = [
    "ape_accounts",
    "ape_cache",
    "ape_compile",
    "ape_console",
    "ape_ethereum",
    "ape_node",
    "ape_init",
    "ape_networks",
    "ape_plugins",
    "ape_pm",
    "ape_run",
    "ape_test",
]
# Hardcoded for performance reasons. Functionality in plugins commands
# and functions won't use this; they use GitHub to check directly.
# This hardcoded list is useful for `ape --help`. If ApeWorX adds a new
# trusted plugin, it should be added to this list; else it will show
# as 3rd-party in `ape --help`.
TRUSTED_PLUGINS = [
    "addressbook",
    "alchemy",
    "arbitrum",
    "avalanche",
    "aws",
    "base",
    "blast",
    "blockscout",
    "bsc",
    "cairo",
    "chainstack",
    "ens",
    "etherscan",
    "fantom",
    "farcaster",
    "flashbots",
    "foundry",
    "frame",
    "ganache",
    "hardhat",
    "infura",
    "ledger",
    "notebook",
    "optimism",
    "polygon",
    "polygon-zkevm",
    "safe",
    "solidity",
    "template",
    "tenderly",
    "titanoboa",
    "tokens",
    "trezor",
    "vyper",
]


def clean_plugin_name(name: str) -> str:
    return name.replace("_", "-").replace("ape-", "")


def get_plugin_dists():
    return _filter_plugins_from_dists(_get_distributions())


class OutputFormat(str, Enum):
    DEFAULT = "default"
    PREFIXED = "prefixed"
    FREEZE = "freeze"


def _filter_plugins_from_dists(dists: Iterable) -> Iterator[str]:
    for dist in dists:
        if name := getattr(dist, "name", ""):
            # Python 3.10 or greater.
            if name.startswith("ape-"):
                yield name

        elif metadata := getattr(dist, "metadata", {}):
            # Python 3.9.
            name = metadata.get("Name", "")
            if name.startswith("ape-"):
                yield name


class ApeVersion:
    def __str__(self) -> str:
        return str(self.version)

    def __getitem__(self, item):
        return str(self)[item]

    @cached_property
    def version(self) -> Version:
        return Version(ape_version_str.split("dev")[0].rstrip("."))

    @property
    def major(self) -> int:
        return self.version.major

    @property
    def minor(self) -> int:
        return self.version.minor

    @property
    def is_pre_one(self) -> bool:
        return self.major == 0

    @cached_property
    def version_range(self) -> str:
        return (
            f">=0.{self.minor},<0.{self.minor + 1}"
            if self.major == 0
            else f">={self.major},<{self.major + 1}"
        )

    @property
    def base(self) -> str:
        return f"0.{self.minor}.0" if self.major == 0 else f"{self.major}.0.0"

    @cached_property
    def next_version_range(self) -> str:
        return (
            f">=0.{self.minor + 1},<0.{self.minor + 2}"
            if self.version.major == 0
            else f">={self.major + 1},<{self.major + 2}"
        )

    @cached_property
    def previous_version_range(self) -> str:
        return (
            f">=0.{self.minor - 2},<0.{self.minor - 1}"
            if self.version.major == 0
            else f">={self.major - 2},<{self.major - 1}"
        )

    def would_get_downgraded(self, plugin_version_str: str) -> bool:
        spec_set = SpecifierSet(plugin_version_str)
        for spec in spec_set:
            spec_version = Version(spec.version)
            if spec.operator in ("==", "<", "<=") and (
                (self.is_pre_one and spec_version.major < self.major)
                or (self.is_pre_one and spec_version.minor < self.minor)
            ):
                return True

        return False


ape_version = ApeVersion()


class PluginType(Enum):
    CORE = "core"
    """
    Plugins that ship with the core product.
    """

    INSTALLED = "installed"
    """
    Plugins that are installed (packages).
    """

    THIRD_PARTY = "third-party"
    """
    Plugins that are installed that are not maintained by a trusted source.
    """

    AVAILABLE = "available"
    """
    Plugins that are available to install from a trusted-source.
    """


class PluginMetadataList(BaseModel):
    """
    Metadata per plugin type, including information for all plugins.
    """

    core: "PluginGroup"
    available: "PluginGroup"
    installed: "PluginGroup"
    third_party: "PluginGroup"

    @classmethod
    def load(cls, plugin_manager: "PluginManager", include_available: bool = True):
        plugins = plugin_manager.registered_plugins
        if include_available:
            plugins = plugins.union(github_client.available_plugins)

        return cls.from_package_names(plugins, include_available=include_available)

    @classmethod
    def from_package_names(
        cls,
        packages: Iterable[str],
        include_available: bool = True,
        trusted_list: Optional[Iterable] = None,
    ) -> "PluginMetadataList":
        PluginMetadataList.model_rebuild()
        core = PluginGroup(plugin_type=PluginType.CORE)
        available = PluginGroup(plugin_type=PluginType.AVAILABLE)
        installed = PluginGroup(plugin_type=PluginType.INSTALLED)
        third_party = PluginGroup(plugin_type=PluginType.THIRD_PARTY)
        for package_id in packages:
            parts = package_id.split("==")
            name = parts[0]
            version = parts[1] if len(parts) == 2 else None
            plugin = PluginMetadata(name=name.strip(), version=version)
            if plugin.in_core:
                core.plugins[name] = plugin
                continue

            # perf: only check these once.
            is_installed = plugin.is_installed
            is_trusted = plugin.check_trusted(use_web=False, trusted_list=trusted_list)
            is_available = include_available and plugin.is_available

            if include_available and is_available and not is_installed:
                available.plugins[name] = plugin
            elif is_installed and not plugin.in_core and is_trusted:
                installed.plugins[name] = plugin
            elif is_installed:
                third_party.plugins[name] = plugin
            else:
                logger.error(f"'{plugin.name}' is not a plugin.")

        return cls(core=core, available=available, installed=installed, third_party=third_party)

    def __str__(self) -> str:
        return self.to_str()

    def to_str(
        self,
        include: Optional[Sequence[PluginType]] = None,
        include_version: bool = True,
        output_format: OutputFormat = OutputFormat.DEFAULT,
    ) -> str:
        representation = ApePluginsRepr(self, include=include, output_format=output_format)
        return str(representation)

    @property
    def all_plugins(self) -> Iterator["PluginMetadata"]:
        yield from self.core.plugins.values()
        yield from self.available.plugins.values()
        yield from self.installed.plugins.values()
        yield from self.third_party.plugins.values()

    def get_plugin(self, name: str, check_available: bool = True) -> Optional["PluginMetadata"]:
        name = name if name.startswith("ape_") else f"ape_{name}"
        if name in self.core.plugins:
            return self.core.plugins[name]
        elif name in self.installed.plugins:
            return self.installed.plugins[name]
        elif name in self.third_party.plugins:
            return self.third_party.plugins[name]
        elif check_available and name in self.available.plugins:
            return self.available.plugins[name]

        return None


def _get_available_plugins():
    # NOTE: Wrapped in a method so can GitHub HTTP can be avoided in tests.
    return github_client.available_plugins


class PluginMetadata(BaseInterfaceModel):
    """
    An encapsulation of a request to install a particular plugin.
    """

    name: str
    """The name of the plugin, such as ``trezor``."""

    version: Optional[str] = None
    """The version requested, if there is one."""

    pip_command: list[str] = PIP_COMMAND
    """
    The pip base command to use.
    (NOTE: is a field mainly for testing purposes).
    """

    @model_validator(mode="before")
    @classmethod
    def validate_name(cls, values):
        if "name" not in values:
            raise ValueError("'name' required.")

        name = values["name"]
        version = values.get("version")

        if name.startswith("git+"):
            version = name
            name = (
                urlparse(version.replace("git+", ""))
                .path.split(".git")[0]
                .split("/")[-1]
                .replace("ape-", "")
            )

        if version and version.startswith("git+"):
            if f"ape-{name}" not in version:
                # Just some small validation so you can't put a repo
                # that isn't this plugin here. NOTE: Forks should still work.
                raise ValueError("Plugin mismatch with remote git version.")

        elif not version:
            # Only check name for version constraint if not in version.
            # NOTE: This happens when using the CLI to provide version constraints.
            for constraint in ("==", "<=", ">=", "@git+"):
                # Version constraint is part of name field.
                if constraint not in name:
                    continue

                # Constraint found.
                name, version = _split_name_and_version(name)
                break

        pip_cmd = values.get("pip_command", PIP_COMMAND)
        return {"name": clean_plugin_name(name), "version": version, "pip_command": pip_cmd}

    @cached_property
    def package_name(self) -> str:
        """
        Like 'ape-plugin'; the name of the package on PyPI.
        """

        return f"ape-{self.name}"

    @cached_property
    def module_name(self) -> str:
        """
        Like 'ape_plugin' or the name you use when importing.
        """

        return f"ape_{self.name.replace('-', '_')}"

    @cached_property
    def current_version(self) -> Optional[str]:
        """
        The version currently installed if there is one.
        """

        return get_package_version(self.package_name)

    @property
    def install_str(self) -> str:
        """
        The strings you pass to ``pip`` to make the install request,
        such as ``ape-trezor==0.4.0``.
        """

        if self.version and self.version.startswith("git+"):
            # If the version is a remote, you do `pip install git+http://github...`
            return self.version

        # `pip install ape-plugin`
        # `pip install ape-plugin==version`.
        # `pip install "ape-plugin>=0.6,<0.7"`

        version = self.version
        if version:
            if not any(x in version for x in ("=", "<", ">")):
                version = f"=={version}"

            # Validate we are not attempting to install a plugin
            # that would change the core-Ape version.
            if ape_version.would_get_downgraded(version):
                raise PluginVersionError(
                    "install", "Doing so will downgrade Ape's version.", "Downgrade Ape first."
                )

        elif not version:
            # When not specifying the version, use a default one that
            # won't dramatically change Ape's version.
            version = ape_version.version_range

        return f"{self.package_name}{version}" if version else self.package_name

    @property
    def can_install(self) -> bool:
        """
        ``True`` if the plugin is available and the requested version differs
        from the installed one.  **NOTE**: Is always ``True`` when the plugin
        is not installed.
        """

        requesting_different_version = (
            self.version is not None and self.version != self.current_version
        )
        return not self.is_installed or requesting_different_version

    @property
    def in_core(self) -> bool:
        """
        ``True`` if the plugin is part of the set of core plugins that
        ship with Ape.
        """

        return self.module_name.strip() in CORE_PLUGINS

    @property
    def is_installed(self) -> bool:
        """
        ``True`` if the plugin is installed in the current Python environment.
        """
        return self.check_installed()

    @property
    def is_third_party(self) -> bool:
        """
        ``True`` when is an installed plugin that is not from ApeWorX.
        """
        return self.is_installed and not self.is_trusted

    @cached_property
    def is_trusted(self) -> bool:
        """
        ``True`` when is a plugin from ApeWorX.
        """
        return self.check_trusted()

    @property
    def is_available(self) -> bool:
        """
        Whether the plugin is maintained by the ApeWorX organization.
        """
        return self.module_name in _get_available_plugins()

    def __str__(self) -> str:
        """
        A string like ``trezor==0.4.0``.
        """
        if self.version and self.version.startswith("git+"):
            return self.version

        version_key = f"=={self.version}" if self.version and self.version[0].isnumeric() else ""
        return f"{self.name}{version_key}"

    def check_installed(self, use_cache: bool = True) -> bool:
        if not use_cache:
            _get_distributions.cache_clear()

        return any(n == self.package_name for n in get_plugin_dists())

    def check_trusted(self, use_web: bool = True, trusted_list: Optional[Iterable] = None) -> bool:
        if use_web:
            return self.is_available

        # Sometimes (such as for --help commands), it is better
        # to not check GitHub to see if the plugin is trusted.
        trusted_list = trusted_list or TRUSTED_PLUGINS
        return self.name in trusted_list

    def prepare_package_manager_args(
        self,
        verb: str,
        python_location: Optional[str] = None,
        extra_args: Optional[Iterable[str]] = None,
    ) -> list[str]:
        """
        Build command arguments for pip or uv package managers.

        Usage:
            args = prepare_package_manager_args("install", "/path/to/python", ["--user"])
        """
        extra_args = list(extra_args) if extra_args else []
        command = list(self.pip_command)

        if command[0] == "uv":
            return (
                [*command, verb, "--python", python_location] + extra_args
                if python_location
                else [*command, verb, *extra_args]
            )

        return (
            [*command, "--python", python_location, verb] + extra_args
            if python_location
            else [*command, verb, *extra_args]
        )

    def _prepare_install(
        self,
        upgrade: bool = False,
        skip_confirmation: bool = False,
        python_location: Optional[str] = None,
    ) -> Optional[dict[str, Any]]:
        # NOTE: Internal and only meant to be called by the CLI.
        if self.in_core:
            logger.error(f"Cannot install core 'ape' plugin '{self.name}'.")
            return None

        elif self.version is not None and upgrade:
            logger.error(
                f"Cannot use '--upgrade' option when specifying a version for plugin '{self.name}'."
            )
            return None

        # if plugin is installed but not trusted. It must be a third party
        elif self.is_third_party:
            logger.warning(f"Plugin '{self.name}' is not an trusted plugin.")

        result_handler = ModifyPluginResultHandler(self)

        pip_arguments = self.prepare_package_manager_args(
            verb="install", python_location=python_location
        )

        if upgrade:
            logger.info(f"Upgrading '{self.name}' plugin ...")

            # NOTE: A simple --upgrade flag may upgrade the plugin
            # to a version outside Core Ape's. Thus, we handle it
            # with a version-specifier instead.
            pip_arguments.extend(
                ("--upgrade", f"{self.package_name}{ape_version.version_range}", "--quiet")
            )
            version_before = self.current_version
            return {
                "args": pip_arguments,
                "version_before": version_before,
                "result_handler": result_handler,
            }

        elif self.can_install and (
            self.is_available
            or skip_confirmation
            or click.confirm(f"Install the '{self.name}' plugin?")
        ):
            logger.info(f"Installing '{self}' plugin ...")
            pip_arguments.extend((self.install_str, "--quiet"))
            return {"args": pip_arguments, "result_handler": result_handler}

        else:
            logger.warning(
                f"'{self.name}' is already installed. Did you mean to include '--upgrade'?"
            )
            return None

    def _get_uninstall_args(self, python_location: Optional[str]) -> list[str]:
        arguments = self.prepare_package_manager_args(
            verb="uninstall", python_location=python_location
        )

        if self.pip_command[0] != "uv":
            arguments.append("-y")

        arguments.extend((self.package_name, "--quiet"))
        return arguments


class ModifyPluginResultHandler:
    def __init__(self, plugin: PluginMetadata):
        self._plugin = plugin

    def handle_install_result(self, result: int) -> bool:
        if not self._plugin.check_installed(use_cache=False):
            self._log_modify_failed("install")
            return False
        elif result != 0:
            self._log_errors_occurred("installing")
            return False
        else:
            plugin_id = self._plugin.name
            version = self._plugin.version
            if version:
                # Sometimes, like in editable mode, the version is missing here.
                plugin_id = f"{plugin_id}=={version}"

            logger.success(f"Plugin '{plugin_id}' has been installed.")
            return True

    def handle_upgrade_result(self, result: int, version_before: str) -> bool:
        if result != 0:
            self._log_errors_occurred("upgrading")
            return False

        version_now = self._plugin.version
        if version_now is not None and version_before == version_now:
            logger.info(f"'{self._plugin.name}' already has version '{version_now}'.")
            return True

        elif self._plugin.version:
            logger.success(
                f"Plugin '{self._plugin.name}' has been upgraded to version {self._plugin.version}."
            )
            return True

        else:
            # The process was successful but there is still no pip freeze version.
            # This may happen when installing things from GitHub.
            return True

    def handle_uninstall_result(self, result) -> bool:
        if self._plugin.check_installed(use_cache=False):
            self._log_modify_failed("uninstall")
            return False
        elif result != 0:
            self._log_errors_occurred("uninstalling")
            return False
        else:
            logger.success(f"Plugin '{self._plugin.name}' has been uninstalled.")
            return True

    def _log_errors_occurred(self, verb: str):
        logger.error(f"Errors occurred when {verb} '{self._plugin}'.")

    def _log_modify_failed(self, verb: str):
        logger.error(f"Failed to {verb} plugin '{self._plugin}.")


def _split_name_and_version(value: str) -> tuple[str, Optional[str]]:
    if "@" in value:
        parts = [x for x in value.split("@") if x]
        return parts[0], "@".join(parts[1:])

    if not (chars := [c for c in ("=", "<", ">") if c in value]):
        return value, None

    index = min(value.index(c) for c in chars)
    return value[:index], value[index:]


class PluginGroup(BaseModel):
    """
    A group of plugin metadata by type.
    """

    plugin_type: PluginType
    plugins: dict[str, PluginMetadata] = {}

    def __bool__(self) -> bool:
        return len(self.plugins) > 0

    @log_instead_of_fail(default="<PluginGroup>")
    def __repr__(self) -> str:
        return f"<{self.name} Plugins Group>"

    def __str__(self) -> str:
        return self.to_str()

    def __iter__(self) -> Iterator[str]:  # type: ignore
        yield from self.plugins

    @field_validator("plugin_type", mode="before")
    @classmethod
    def validate_plugin_type(cls, value):
        return PluginType(value) if isinstance(value, str) else value

    @property
    def plugin_type_str(self) -> str:
        return getattr(self.plugin_type, "value", str(self.plugin_type))

    @property
    def name(self) -> str:
        return self.plugin_type_str.capitalize()

    @property
    def plugin_names(self) -> list[str]:
        return [x.name for x in self.plugins.values()]

    def to_str(
        self,
        max_length: Optional[int] = None,
        include_version: bool = True,
        output_format: Optional[OutputFormat] = OutputFormat.DEFAULT,
    ) -> str:
        output_format = output_format or OutputFormat.DEFAULT
        if output_format in (OutputFormat.DEFAULT, OutputFormat.PREFIXED):
            return self._get_default_formatted_str(
                max_length=max_length,
                include_version=include_version,
                include_prefix=f"{output_format}" == OutputFormat.PREFIXED,
            )

        # Freeze format.
        return self._get_freeze_formatted_str(
            max_length=max_length, include_version=include_version
        )

    def _get_default_formatted_str(
        self,
        max_length: Optional[int] = None,
        include_version: bool = True,
        include_prefix: bool = False,
    ) -> str:
        title = f"{self.name} Plugins"
        if len(self.plugins) <= 0:
            return title

        lines = [title]
        max_length = self.max_name_length if max_length is None else max_length
        plugins_sorted = sorted(self.plugins.values(), key=lambda p: p.name)
        for plugin in plugins_sorted:
            line = plugin.package_name if include_prefix else plugin.name
            if include_version:
                version = plugin.version or get_package_version(plugin.package_name)
                if version:
                    spacing = (max_length - len(line)) + 4
                    line = f"{line}{spacing * ' '}{version}"

            lines.append(f"  {line}")  # Indent.

        return "\n".join(lines)

    def _get_freeze_formatted_str(
        self,
        max_length: Optional[int] = None,
        include_version: bool = True,
        include_prefix: bool = False,
    ) -> str:
        lines = []
        for plugin in sorted(self.plugins.values(), key=lambda p: p.name):
            line = plugin.package_name
            if include_version:
                version = plugin.version or get_package_version(plugin.package_name)
                line = f"{line}=={version}"

            lines.append(line)

        return "\n".join(lines)

    @property
    def max_name_length(self) -> int:
        if not self.plugins:
            return 0

        return max(len(x) for x in self.plugin_names)


class ApePluginsRepr:
    """
    A str-builder for all installed Ape plugins.
    """

    def __init__(
        self,
        metadata: PluginMetadataList,
        include: Optional[Sequence[PluginType]] = None,
        include_version: bool = True,
        output_format: Optional[OutputFormat] = None,
    ):
        self.include = include or (PluginType.INSTALLED, PluginType.THIRD_PARTY)
        self.metadata = metadata
        self.include_version = include_version
        self.output_format = output_format

    @log_instead_of_fail(default="<ApePluginsRepr>")
    def __repr__(self) -> str:
        to_display_str = ", ".join([x.value for x in self.include])
        return f"<PluginMap to_display='{to_display_str}'>"

    def __str__(self) -> str:
        sections = []

        if PluginType.CORE in self.include and self.metadata.core:
            sections.append(self.metadata.core)
        if PluginType.INSTALLED in self.include and self.metadata.installed:
            sections.append(self.metadata.installed)
        if PluginType.THIRD_PARTY in self.include and self.metadata.third_party:
            sections.append(self.metadata.third_party)
        if PluginType.AVAILABLE in self.include and self.metadata.available:
            sections.append(self.metadata.available)

        if not sections:
            return ""

        # Use a single max length for all the sections.
        max_length = max(x.max_name_length for x in sections)

        version_skips = (PluginType.CORE, PluginType.AVAILABLE)

        def include_version(section):
            return section.plugin_type not in version_skips if self.include_version else False

        formatted_sections = [
            x.to_str(
                max_length=max_length,
                include_version=include_version(x),
                output_format=self.output_format,
            )
            for x in sections
        ]
        return "\n\n".join(formatted_sections)



from importlib import import_module

from ape.plugins import AccountPlugin, register


@register(AccountPlugin)
def account_types():
    from ape_accounts.accounts import AccountContainer, KeyfileAccount

    return AccountContainer, KeyfileAccount


def __getattr__(name: str):
    return getattr(import_module("ape_accounts.accounts"), name)


__all__ = [
    "AccountContainer",
    "KeyfileAccount",
    "generate_account",
    "import_account_from_mnemonic",
    "import_account_from_private_key",
]



import faulthandler
import inspect
import logging
import sys
from functools import cached_property
from importlib import import_module
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader
from os import environ
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import click

from ape.cli.commands import ConnectedProviderCommand
from ape.cli.options import ape_cli_context, project_option

if TYPE_CHECKING:
    from IPython.terminal.ipapp import Config as IPythonConfig

    from ape.managers.project import ProjectManager

CONSOLE_EXTRAS_FILENAME = "ape_console_extras.py"


def _code_callback(ctx, param, value) -> list[str]:
    if not value:
        return value

    # NOTE: newlines are escaped in code automatically, so we
    #   need to de-escape them. Any actually escaped newlines
    #   will still be escaped.
    value = value.replace("\\n", "\n").replace("\\t", "\t").replace("\\b", "\b")

    return value.splitlines()


@click.command(
    cls=ConnectedProviderCommand,
    short_help="Load the console",
    context_settings=dict(ignore_unknown_options=True),
)
@ape_cli_context()
@project_option(hidden=True, type=Path)  # Hidden as mostly used for test purposes.
@click.option("-c", "--code", help="Program passed in as a string", callback=_code_callback)
def cli(cli_ctx, project, code):
    """Opens a console for the local project."""
    verbose = cli_ctx.logger.level == logging.DEBUG
    return console(project=project, verbose=verbose, code=code)


def import_extras_file(file_path) -> ModuleType:
    """Import a module"""
    loader = SourceFileLoader(file_path.name[:-3], str(file_path))
    spec = spec_from_loader(loader.name, loader)

    assert spec is not None

    module = module_from_spec(spec)
    loader.exec_module(module)

    return module


class ApeConsoleNamespace(dict):
    def __init__(self, **kwargs):
        # Initialize the dictionary with provided keyword arguments
        project = kwargs.get("project", self._ape.project)
        kwargs["project"] = self._ape.Project(project) if isinstance(project, Path) else project
        super().__init__(**kwargs)

    def __getitem__(self, key: str):
        # First, attempt to retrieve the key from the dictionary itself
        if super().__contains__(key):
            return super().__getitem__(key)

        # Custom behavior for "ape" key
        if key == "ape":
            res = self._ape
            self[key] = res  # Cache the result
            return res

        # Attempt to get the key from extras
        try:
            res = self._get_extra(key)
        except KeyError:
            pass
        else:
            self[key] = res  # Cache the result
            return res

        # Attempt to retrieve the key from the Ape module.
        try:
            res = self._get_from_ape(key)
        except AttributeError:
            raise KeyError(key)

        # Cache the result and return
        self[key] = res
        return res

    def __setitem__(self, key, value):
        # Override to set items directly into the dictionary
        super().__setitem__(key, value)

    def __contains__(self, item: str) -> bool:  # type: ignore
        return self.get(item) is not None

    def update(self, mapping, **kwargs) -> None:  # type: ignore
        # Override to update the dictionary directly
        super().update(mapping, **kwargs)

    @property
    def _ape(self) -> ModuleType:
        return import_module("ape")

    @cached_property
    def _local_path(self) -> Path:
        return self["project"].path.joinpath(CONSOLE_EXTRAS_FILENAME)

    @cached_property
    def _global_path(self) -> Path:
        return self._ape.config.DATA_FOLDER.joinpath(CONSOLE_EXTRAS_FILENAME)

    @cached_property
    def _local_extras(self) -> dict:
        return self._load_extras_file(self._local_path)

    @cached_property
    def _global_extras(self) -> dict:
        return self._load_extras_file(self._global_path)

    def get(self, key: str, default: Optional[Any] = None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def _get_extra(self, key: str):
        try:
            return self._local_extras[key]
        except KeyError:
            return self._global_extras[key]

    def _get_from_ape(self, key: str):
        return getattr(self._ape, key)

    def _load_extras_file(self, extras_file: Path) -> dict:
        if not extras_file.is_file():
            return {}

        module = import_extras_file(extras_file)
        ape_init_extras = getattr(module, "ape_init_extras", None)
        all_extras: dict = {}

        if ape_init_extras is not None:
            func_spec = inspect.getfullargspec(ape_init_extras)
            init_kwargs: dict[str, Any] = {k: self._get_from_ape(k) for k in func_spec.args}
            extras = ape_init_extras(**init_kwargs)

            if isinstance(extras, dict):
                all_extras.update(extras)

        all_extras.update({k: getattr(module, k) for k in dir(module) if k not in all_extras})
        return all_extras


def console(
    project: Optional[Union["ProjectManager", Path]] = None,
    verbose: bool = False,
    extra_locals: Optional[dict] = None,
    embed: bool = False,
    code: Optional[list[str]] = None,
):
    import IPython
    from IPython.terminal.ipapp import Config as IPythonConfig

    from ape.utils.misc import _python_version
    from ape.version import version as ape_version

    extra_locals = extra_locals or {}
    if project is None:
        from ape.utils.basemodel import ManagerAccessMixin

        project = ManagerAccessMixin.local_project

    else:
        extra_locals["project"] = project

    project_path: Path = project if isinstance(project, Path) else project.path
    banner = ""
    if verbose:
        banner = f"""
   Python:  {_python_version}
  IPython:  {IPython.__version__}
      Ape:  {ape_version}
  Project:  {project_path}

    Are you ready to Ape, anon?
    """

        if not environ.get("APE_TESTING"):
            faulthandler.enable()  # NOTE: In case we segfault

    # Allows modules relative to the project.
    sys.path.insert(0, f"{project_path}")

    ipy_config = IPythonConfig()
    ape_testing = environ.get("APE_TESTING")
    if ape_testing:
        ipy_config.HistoryManager.enabled = False

        # Required for click.testing.CliRunner support.
        embed = True

    namespace = _create_namespace(**extra_locals)
    _launch_console(namespace, ipy_config, embed, banner, code=code)


def _create_namespace(**values) -> dict:
    # Abstracted for testing purposes.
    return ApeConsoleNamespace(**values)


def _launch_console(
    namespace: dict,
    ipy_config: "IPythonConfig",
    embed: bool,
    banner: str,
    code: Optional[list[str]],
):
    import IPython

    from ape_console.config import ConsoleConfig

    ipython_kwargs = {"user_ns": namespace, "config": ipy_config}
    if code:
        _execute_code(code, **ipython_kwargs)
    elif embed:
        IPython.embed(**ipython_kwargs, colors="Neutral", banner1=banner)
    else:
        ipy_config.TerminalInteractiveShell.colors = "Neutral"
        ipy_config.TerminalInteractiveShell.banner1 = banner
        console_config = cast(ConsoleConfig, namespace["ape"].config.get_config("console"))
        ipy_config.InteractiveShellApp.extensions.append("ape_console.plugin")
        if console_config.plugins:
            ipy_config.InteractiveShellApp.extensions.extend(console_config.plugins)

        IPython.start_ipython(**ipython_kwargs, argv=())


def _execute_code(code: list[str], **ipython_kwargs):
    from IPython import InteractiveShell

    shell = InteractiveShell.instance(**ipython_kwargs)
    # NOTE: Using `store_history=True` just so the cell IDs are accurate.
    for line in code:
        shell.run_cell(line, store_history=True)



import copy
from pathlib import Path

import pytest

from ape.contracts import ContractContainer
from ape_node.provider import Node
from tests.functional.data.python import TRACE_RESPONSE


@pytest.fixture(scope="session")
def safe_proxy_container(get_contract_type):
    proxy_type = get_contract_type("SafeProxy")
    return ContractContainer(proxy_type)


@pytest.fixture
def parity_trace_response():
    return TRACE_RESPONSE


@pytest.fixture
def contract_with_call_depth_geth(
    owner, geth_provider, get_contract_type, leaf_contract_geth, middle_contract_geth
):
    """
    This contract has methods that make calls to other local contracts
    and is used for any testing that requires nested calls, such as
    call trees or event-name clashes.
    """
    contract = ContractContainer(get_contract_type("ContractA"))
    return owner.deploy(contract, middle_contract_geth, leaf_contract_geth)


@pytest.fixture
def error_contract_geth(owner, error_contract_container, geth_provider):
    _ = geth_provider  # Ensure uses geth
    return owner.deploy(error_contract_container, 1)


@pytest.fixture
def leaf_contract_geth(geth_provider, owner, get_contract_type):
    """
    The last contract called by `contract_with_call_depth`.
    """
    ct = get_contract_type("ContractC")
    return owner.deploy(ContractContainer(ct))


@pytest.fixture
def middle_contract_geth(geth_provider, owner, leaf_contract_geth, get_contract_type):
    """
    The middle contract called by `contract_with_call_depth`.
    """
    ct = get_contract_type("ContractB")
    return owner.deploy(ContractContainer(ct), leaf_contract_geth)


@pytest.fixture
def mock_geth(geth_provider, mock_web3):
    provider = Node(
        name="node",
        network=geth_provider.network,
        provider_settings={},
        data_folder=Path(),
    )
    original_web3 = provider._web3
    provider._web3 = mock_web3
    yield provider
    provider._web3 = original_web3


@pytest.fixture
def geth_receipt(contract_with_call_depth_geth, owner, geth_provider):
    return contract_with_call_depth_geth.methodWithoutArguments(sender=owner)


@pytest.fixture
def geth_vyper_receipt(geth_vyper_contract, owner):
    return geth_vyper_contract.setNumber(44, sender=owner)


@pytest.fixture
def custom_network_connection(
    geth_provider,
    ethereum,
    project,
    custom_network_name_0,
    custom_networks_config_dict,
    networks,
):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["chain_id"] = geth_provider.chain_id

    config = {
        ethereum.name: {custom_network_name_0: {"default_transaction_type": 0}},
        geth_provider.name: {ethereum.name: {custom_network_name_0: {"uri": geth_provider.uri}}},
        **data,
    }
    actual = geth_provider.network
    with project.temp_config(**config):
        geth_provider.network = ethereum.apenet
        try:
            with networks.ethereum.apenet.use_provider("node"):
                yield

        finally:
            geth_provider.network = actual



import pytest

from ape.exceptions import ContractLogicError, OutOfGasError
from tests.conftest import geth_process_test


@geth_process_test
def test_contract_interaction(geth_provider, geth_account, geth_contract, mocker):
    # Spy on the estimate_gas RPC method.
    estimate_gas_spy = mocker.spy(geth_provider.web3.eth, "estimate_gas")

    # Check what max gas is before transacting.
    max_gas = geth_provider.max_gas

    # Invoke a method from a contract via transacting.
    receipt = geth_contract.setNumber(102, sender=geth_account)

    # Verify values from the receipt.
    assert not receipt.failed
    assert receipt.receiver == geth_contract.address
    assert receipt.gas_used < receipt.gas_limit
    assert receipt.gas_limit == max_gas

    # Show contract state changed.
    assert geth_contract.myNumber() == 102

    # Verify the estimate gas RPC was not used (since we are using max_gas).
    assert estimate_gas_spy.call_count == 0


@geth_process_test
def test_contract_call_show_trace(geth_contract, geth_account):
    """
    Show the `show_trace=True` does not corrupt the value.
    Note: The provider uses `debug_traceCall` to get the result instead of
    `eth_call`.
    """
    geth_contract.setNumber(203, sender=geth_account)
    actual = geth_contract.myNumber(show_trace=True)
    assert actual == 203


@geth_process_test
def test_tx_revert(accounts, not_owner, geth_contract):
    # 'sender' is not the owner so it will revert (with a message)
    with pytest.raises(ContractLogicError, match="!authorized") as err:
        geth_contract.setNumber(5, sender=not_owner)

    assert err.value.txn is not None


@geth_process_test
def test_revert_no_message(accounts, geth_contract, geth_account):
    # The Contract raises empty revert when setting number to 5.
    expected = "Transaction failed."  # Default message
    with pytest.raises(ContractLogicError, match=expected) as err:
        geth_contract.setNumber(5, sender=geth_account)

    assert err.value.txn is not None


@geth_process_test
def test_revert_custom_error(error_contract_geth, geth_second_account):
    contract = error_contract_geth
    with pytest.raises(contract.Unauthorized) as err:
        contract.withdraw(sender=geth_second_account)

    assert err.value.txn is not None
    assert err.value.inputs == {"addr": geth_second_account.address, "counter": 123}


@geth_process_test
def test_revert_custom_error_deploy(error_contract_container, geth_account, chain, geth_provider):
    with pytest.raises(Exception) as err:
        geth_account.deploy(error_contract_container, 0)

    err_cls = err.value
    assert isinstance(err_cls, ContractLogicError)
    if err_cls.address:
        contract = chain.contracts.instance_at(err_cls.address)

        # Ensure it is the custom error.
        assert isinstance(err_cls, contract.OtherError)

    else:
        # skip this test - still covered in reverts() tests anyway.
        return


@geth_process_test
def test_revert_out_of_gas_error(geth_account, geth_second_account, geth_provider):
    """
    Attempt to transact with not quite enough gas. We should get an error saying
    we ran out of gas.
    """
    with pytest.raises(OutOfGasError) as err:
        geth_account.transfer(geth_second_account, 1, gas_limit=1)

    assert err.value.txn is not None


@geth_process_test
def test_revert_out_of_gas_error_allow(geth_account, geth_second_account, geth_provider):
    tx = geth_account.transfer(geth_second_account, 1, gas_limit=1, raise_on_revert=False)
    assert tx.failed
    assert isinstance(tx.error, OutOfGasError)


@geth_process_test
def test_revert_allow(accounts, geth_contract):
    not_owner = accounts[0]

    # 'sender' is not the owner so it will revert (with a message)
    receipt = geth_contract.setNumber(100199, sender=not_owner, raise_on_revert=False)
    assert receipt.error is not None
    assert str(receipt.error) == "!authorized"

    # Ensure this also works for calls.
    geth_contract.setNumber.call(5, raise_on_revert=False)



from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from ape_node.provider import EthereumNodeConfig

    return EthereumNodeConfig


@plugins.register(plugins.ProviderPlugin)
def providers():
    from ape_node.provider import EthereumNetworkConfig, GethDev, Node

    networks_dict = EthereumNetworkConfig().model_dump()
    networks_dict.pop("local")
    for network_name in networks_dict:
        yield "ethereum", network_name, Node

    yield "ethereum", "local", GethDev


@plugins.register(plugins.QueryPlugin)
def query_engines():
    from ape_node.query import OtterscanQueryEngine

    yield OtterscanQueryEngine


def __getattr__(name: str):
    if name == "OtterscanQueryEngine":
        from ape_node.query import OtterscanQueryEngine

        return OtterscanQueryEngine

    import ape_node.provider as module

    return getattr(module, name)


__all__ = [
    "EthereumNetworkConfig",
    "EthereumNodeConfig",
    "GethDev",
    "Node",
    "OtterscanQueryEngine",
]
